# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

"""
[Name] Harvester

[Description]
Harvester is use as a replacement of the Dewar sample storage
This hardware object is use in couple with a Sample changer.
Sample changer load the sample from Harvester instead of dewar.
It is compatible with the Crystal Direct Harvester 3.
It has some functionalities, like Harvest Sample, etc....

[Commands]

 - getSampleList : Get list of available sample from Harvester
 - Harvest : Harvest sample make it ready to load

 [Example xml file:]
    <object class="Harvester">
        <username>harvester</username>
        <exporter_address>wid30harvest:9001</exporter_address>
    </object>
-----------------------------------------------------------------
"""

from __future__ import annotations

import logging
from typing import List, Optional, Union

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObject


class HarvesterState:
    """
    Enumeration of Harvester states
    """

    Unknown = 0
    Initializing = 1
    Ready = 2
    Harvested = 3
    Running = 4
    Harvesting = 5
    ContinueHarvesting = 6
    # Disabled = 7
    # Running = 8
    # StandBy = 9
    # Alarm = 10
    # Fault = 11

    STATE_DESC = {
        Initializing: "Initializing",
        Ready: "Ready",
        Harvested: "Waiting Sample Transfer",
        Running: "Running",
        Harvesting: "Harvesting 1 Crystals",
        ContinueHarvesting: "Finishing Harvesting",
    }

    @staticmethod
    def tostring(state):
        return HarvesterState.STATE_DESC.get(state, "Unknown")


class Harvester(HardwareObject):
    """
    Harvester functionality

    The Harvester Class consists of methods that execute exporter commands
    this class communicate with the Crystal Direct Harvester Machine

    """

    __TYPE__ = "Harvester"

    def __init__(self, name):
        super().__init__(name)
        self.timeout = 600  # default timeout

        # Internal variables -----------
        self.calibration_state = False

    def init(self):
        """Init"""
        self.exporter_addr = self.get_property("exporter_address")
        self.crims_upload_url = self.get_property("crims_upload_url")
        self.crims_upload_key = self.get_property("crims_upload_key")

    def set_calibration_state(self, state: bool):
        """Set Calibration state

        Args:
        state (bool) : Whether a calibration procedure is on going
        """

        self.calibration_state = state

    def _wait_ready(self, timeout: Union[float, None] = None):
        """Wait Harvester to be ready

        Args:
        (timeout) : Whether to wait for a amount of time
        timeout is None wait forever, timeout <=0 use default timeout
        """
        if timeout is not None and timeout <= 0:
            timeout = self.timeout

        err_msg = "Timeout waiting for Harvester to be ready"

        with gevent.Timeout(timeout, RuntimeError(err_msg)):
            while not self._ready():
                logging.getLogger("user_level_log").info(
                    "Waiting Harvester to be Ready"
                )
                gevent.sleep(3)

    def _wait_sample_transfer_ready(self, timeout: Union[float, None] = None):
        """Wait Harvester to be ready to transfer a sample

        Args:
        timeout (second) : Whether to wait for a amount of time
        timeout is None wait forever, timeout <=0 use default timeout
        """
        if timeout is not None and timeout <= 0:
            timeout = self.timeout

        err_msg = "Timeout waiting for Harvester to be ready to transfer"

        try:
            with gevent.Timeout(timeout, RuntimeError(err_msg)):
                while not self._ready_to_transfer():
                    logging.getLogger("user_level_log").info(
                        "Waiting Harvester to be ready to transfer for 10 minutes"
                    )
                    gevent.sleep(3)
        except RuntimeError as exc:
            # In case of timeout we as abort, park and trash
            self.abort()
            self.park()
            self._wait_ready(self.timeout)
            logging.getLogger("user_level_log").info("Trash current Sample")
            self.trash_sample()
            raise RuntimeError("Harvester failed to become ready in time") from exc

    def _execute_cmd_exporter(self, cmd, *args, **kwargs):
        """Exporter Command implementation

        Args:
        cmd (string) : command type
        args, kwargs (string): commands arguments, and  command or attribute

        return : respond
        """
        ret = None
        if args:
            args_str = "%s" % "\t".join(map(str, args))
        if kwargs.pop("command", None):
            exp_cmd = self.add_command(
                {
                    "type": "exporter",
                    "exporter_address": self.exporter_addr,
                    "name": "%s" % cmd,
                },
                "%s" % cmd,
            )
            if args:
                ret = exp_cmd(args_str)
            else:
                ret = exp_cmd()
        if kwargs.pop("attribute", None):
            exp_attr = self.add_channel(
                {
                    "type": "exporter",
                    "exporter_address": self.exporter_addr,
                    "name": "%s" % cmd,
                },
                "%s" % cmd[3:],
            )
            if cmd.startswith("get"):
                return exp_attr.get_value()
            if cmd.startswith("set"):
                ret = exp_attr.set_value(args_str)

        return ret

    # ---------------------- State --------------------------------

    def get_state(self) -> str:
        """Get the Harvester State

        Return (str):  state "Ready, Running etc.."
        """
        return self._execute_cmd_exporter("getState", attribute=True)

    def get_status(self) -> str:
        """Get the Harvester Status

        Return (str):  Status
        """
        return self._execute_cmd_exporter("getStatus", attribute=True)

    def _ready(self) -> str:
        """check whether the Harvester is READY

        Return (bool):  True if Harvester is Ready otherwise False
        """
        return self._execute_cmd_exporter("getState", attribute=True) == "Ready"

    def _busy(self) -> bool:
        """check whether the Harvester is BUSY

        Return (bool):  True if Harvester is not Ready otherwise False
        """
        return self._execute_cmd_exporter("getState", attribute=True) != "Ready"

    def _ready_to_transfer(self) -> bool:
        """check whether the Harvester is Waiting Sample Transfer

        Return (bool):  True if Harvester is Waiting Sample Transfer otherwise False
        """
        return (
            self._execute_cmd_exporter("getStatus", attribute=True)
            == "Waiting Sample Transfer"
        )

    def get_samples_state(self) -> List[str]:
        """Get the Harvester Samples State

        Return (List):  list of crystal state "waiting_for_transfer, Running etc.."
        """
        return self._execute_cmd_exporter("getSampleStates", attribute=True)

    def get_current_crystal(self) -> str:
        """Get the Harvester current harvested crystal

        Return (str): the crystal uuid
        """
        return self._execute_cmd_exporter("getCurrentSampleID", attribute=True)

    def is_crystal_harvested(self, crystal_uuid: str) -> str:
        """Check Whether if the current crystal is harvested

        args: the crystal uuid

        Return (bool):  True if the crystal is the current harvested crystal
        """
        res = False
        in_list = crystal_uuid in self.get_crystal_uuids()
        if in_list:
            Current_SampleID = self.get_current_crystal()
            if crystal_uuid == Current_SampleID:
                res = True
        return res

    def current_crystal_state(self, crystal_uuid: str) -> str:
        """get current crystal state

        Args:
        state (str) : Crystal uuid

        Return (str):  State of the crystal uuid
        """
        sample_states = self.get_samples_state()
        crystal_uuids = self.get_crystal_uuids()

        for index, x_tal in enumerate(crystal_uuids):
            if crystal_uuid == x_tal:
                return sample_states[index]

        return None

    def check_crystal_state(self, crystal_uuid: str) -> Optional[str]:
        """Check whether if a Crystal is in pending_and_current or not

        Args (str) : Crystal uuid

        Return (str):  status of the crystal_uuid pending / current
        """
        sample_states = self.get_samples_state()
        crystal_uuids = self.get_crystal_uuids()

        for index, x_tal in enumerate(crystal_uuids):
            if crystal_uuid == x_tal and sample_states[index] == "waiting_for_transfer":
                return "pending_and_current"
            elif (
                crystal_uuid != x_tal and sample_states[index] == "waiting_for_transfer"
            ):
                return "pending_not_current"
            else:
                return None

    def get_crystal_uuids(self) -> List[str]:
        """Get the Harvester Sample List uuid

        Return (List):  list of crystal by uuid from the current processing plan"
        """
        harvester_crystal_list = self._execute_cmd_exporter(
            "getSampleList", attribute=True
        )
        return harvester_crystal_list

    def get_sample_names(self) -> List[str]:
        """Get the Harvester Sample List Name

        Return (List):  list of crystal by names from the current processing plan"
        """
        harvester_sample_names = self._execute_cmd_exporter(
            "getSampleNames", attribute=True
        )
        return harvester_sample_names

    def get_crystal_images_urls(self, crystal_uuid: str) -> str:
        """Get the Harvester Sample List Images

        Args (str) : Crystal uuid

        Return (str): image_url from current processing plan for current crystal uuid"
        """
        crystal_images_url = self._execute_cmd_exporter(
            "getImageURL", crystal_uuid, command=True
        )
        return crystal_images_url

    def get_sample_acronyms(self) -> List[str]:
        """Get the Harvester Sample List by Acronyms

        Return (List):  list of crystal by Acronyms from the current processing plan"
        """
        harvester_sample_acronyms = self._execute_cmd_exporter(
            "getSampleAcronyms", attribute=True
        )
        return harvester_sample_acronyms

    # ------------------------------------------------------------------------------------

    def abort(self) -> str:
        """Send Abort command
        Abort any current Harvester Actions
        """
        return self._execute_cmd_exporter("abort", command=True)

    def park(self) -> str:
        """Send Park command
        Park Harvester Actions
        """
        return self._execute_cmd_exporter("park", command=True)

    def harvest_crystal(self, crystal_uuid: str) -> str:
        """Harvester crystal

        Args (str) : Crystal uuid
        """
        try:
            self._execute_cmd_exporter("harvestCrystal", crystal_uuid, command=True)
            return "Crystal Harvested properly"
        except Exception:
            logging.getLogger("user_level_log").warning(
                f"Warning: Could not harvest sample: {crystal_uuid}"
            )
            return "Could not Harvest Crystal"

    def transfer_sample(self) -> None:
        """Transfer the current Harvested Crystal"""
        return self._execute_cmd_exporter("startTransfer", command=True)

    def trash_sample(self):
        """Trash the current Harvested Crystal"""
        return self._execute_cmd_exporter("trashSample", command=True)

    # -----------------------------------------------------------------------------

    def load_plate(self, plate_id: str) -> str:
        """Change Harvester current plate

        Args (str) : Plate ID
        Return (str) : current Plate ID
        """
        self._execute_cmd_exporter("loadPlate", plate_id, command=True)
        return self._execute_cmd_exporter("getPlateID", attribute=True)

    def get_plate_id(self) -> str:
        """get current plate ID

        Args:
        Return (str) : current Plate ID
        """
        return self._execute_cmd_exporter("getPlateID", attribute=True)

    def get_image_target_x(self, crystal_uuid: str) -> float:
        """Get the crystal images position x

        Args (str) : Crystal uuid

        Return (float):  Crystal x coordinate in plate
        """
        return self._execute_cmd_exporter("getImageTargetX", crystal_uuid, command=True)

    def get_image_target_y(self, crystal_uuid: str) -> float:
        """Get the crystal images position Y

        Args (str) : Crystal uuid

        Return (float): Crystal Y coordinate in plate
        """
        return self._execute_cmd_exporter("getImageTargetY", crystal_uuid, command=True)

    def get_room_temperature_mode(self) -> bool:
        """get  RoomTemperature Mode state

        Args (str) : Crystal uuid

        Return (bool):  TemperatureMode , True if Room Temp else False
        """
        return self._execute_cmd_exporter("getRoomTemperatureMode", attribute=True)

    def set_room_temperature_mode(self, value: bool) -> bool:
        """Set Harvester temperature mode

        Args: (bool) set room temperature when true

        Return (bool):  TemperatureMode
        """
        self._execute_cmd_exporter("setRoomTemperatureMode", value, command=True)
        print("setting HA Room temperature to: %s" % value)
        return self.get_room_temperature_mode()

    # -------------------- Calibrate  Drift Shape offset ----------------------------

    def get_last_sample_drift_offset_x(self) -> float:
        """Sample Offset X position when drifted
        Return (float):  last pin drift offset x
        """
        last_sample_drift_offset_x = self._execute_cmd_exporter(
            "getLastSampleDriftOffsetX", attribute=True
        )
        return last_sample_drift_offset_x

    def get_last_sample_drift_offset_y(self) -> float:
        """Sample Offset Y position when drifted
        Return (float):  last pin drift offset y
        """
        last_sample_drift_offset_y = self._execute_cmd_exporter(
            "getLastSampleDriftOffsetY", attribute=True
        )
        return last_sample_drift_offset_y

    def get_last_sample_drift_offset_z(self) -> float:
        """Sample Offset Z position when drifted
        Return (float):  last pin drift offset z
        """
        pin_last_drift_offset_z = self._execute_cmd_exporter(
            "getLastSampleDriftOffsetZ", attribute=True
        )
        return pin_last_drift_offset_z

    # ---------------------- Calibrate Cut Shape offset----------------------------

    def get_last_pin_cut_shape_offset_x(self) -> float:
        """Pin shape Offset x position
        Return (float):  last pin cut shape offset x
        """
        pin_last_cut_shape_offset_x = self._execute_cmd_exporter(
            "getLastSampleCutShapeOffsetX", attribute=True
        )
        return pin_last_cut_shape_offset_x

    def get_last_pin_cut_shape_offset_y(self) -> float:
        """Pin shape Offset Y position
        Return (float):  last pin cut shape offset y
        """
        pin_last_cut_shape_offset_y = self._execute_cmd_exporter(
            "getLastSampleCutShapeOffsetY", attribute=True
        )
        return pin_last_cut_shape_offset_y

    # =============== Pin / Calibration -----------------------------

    def load_calibrated_pin(self) -> None:
        """Load a Pin for Calibration Procedure"""
        return self._execute_cmd_exporter("loadCalibratedPin", command=True)

    def store_calibrated_pin(self, x: float, y: float, z: float) -> None:
        """Store x , y , z offsets position to crystal direct machine
        after calibration procedure

        Args: (float) x, y, z offsets
        """
        return self._execute_cmd_exporter("storePinToBeamOffset", x, y, z, command=True)

    def get_calibrated_pin_offset(self) -> tuple[float]:
        """Get Stored x , y , z offsets position after calibration procedure

        return: (float) x, y, z offsets
        """
        pin_to_beam_offset = self._execute_cmd_exporter(
            "getPinToBeamOffset", command=True
        )
        return pin_to_beam_offset

    def get_number_of_available_pin(self) -> int:
        """Get number of available pin

        return: (Integer)
        """
        return self._execute_cmd_exporter("getNbRemainingPins", command=True)

    def queue_harvest_sample(
        self, sample_loc_str, sample_uuid: str, current_queue_list: list[str]
    ) -> bool:
        """
        While queue execution send harvest request
        current_queue_list : a build representation of the queue
        """
        harvest_res = False
        current_queue_index = None
        try:
            current_queue_index = current_queue_list.index(sample_loc_str)
        except (ValueError, IndexError):
            current_queue_index = None

        wait_before_load = not self.get_room_temperature_mode()

        if self.get_number_of_available_pin() > 0:
            gevent.sleep(2)

            if current_queue_index == 0:
                logging.getLogger("user_level_log").info("Harvesting First Sample")

                harvest_res = self.harvest_sample_before_mount(
                    sample_uuid, wait_before_load
                )
                if harvest_res is False:
                    # if sample could not be Harvest, but no exception is raised, let's skip the sample
                    logging.getLogger("user_level_log").error(
                        "Harvester could not Harvest sample, Stopping queue"
                    )
            else:
                logging.getLogger("user_level_log").info("checking last Harvesting")
                harvest_res = self.harvest_sample_before_mount(
                    sample_uuid, wait_before_load
                )
                if harvest_res is False:
                    # if sample could not be Harvest, but no exception is raised, let's skip the sample
                    logging.getLogger("user_level_log").error(
                        "There is no more Pins in the Harvester, Stopping queue"
                    )
        elif self.get_number_of_available_pin() == 0 and self._ready_to_transfer():
            logging.getLogger("user_level_log").warning(
                "Warning: Harvester pins is approaching to ZERO"
            )
            logging.getLogger("user_level_log").warning(
                "Warning: Mounting last Sample, Queue will stop on next one"
            )
            # in this case we just load the sample that is ready in the Harester
            harvest_res = True
        else:
            # raise Not enough pins available in the pin provider
            logging.getLogger("user_level_log").error(
                "There is no more Pins in the Harvester, Stopping queue"
            )

        return harvest_res

    def queue_harvest_next_sample(self, next_sample_loc_str: str, sample_uuid: str):
        """
        While queue execution send harvest request
        on next sample of the queue list
        """

        if next_sample_loc_str is not None and self.get_number_of_available_pin() > 0:
            logging.getLogger("user_level_log").info("Harvesting Next Sample")

            self._wait_ready(self.timeout)
            self.harvest_sample_before_mount(sample_uuid, False)
        else:
            logging.getLogger("user_level_log").warning(
                "Warning: Could not harvest next sample"
            )

    def harvest_sample_before_mount(
        self, sample_uuid: str, wait_before_load: bool = False
    ) -> bool:
        """send harvest sample command
            Check and set  the current state of the Harvester and the sample before Harvest

        Return (bool): whether the sample has been harvest then mount (True)
        or had and exception (False)
        """
        res = None

        if sample_uuid:
            if self.get_status() == "Ready":
                try:
                    if self.check_crystal_state(sample_uuid) == "pending_not_current":
                        print(self.get_samples_state())
                        logging.getLogger("user_level_log").info(
                            "Harvester:Trashing pending Sample"
                        )
                        self.trash_sample()
                        self._wait_ready(self.timeout)
                    if (
                        self.current_crystal_state(sample_uuid) == "ready_to_execute"
                        or self.current_crystal_state(sample_uuid)
                        == "needs_repositionning"
                    ):
                        logging.getLogger("user_level_log").info("Harvesting started")
                        self.harvest_crystal(sample_uuid)
                        if wait_before_load:
                            self._wait_sample_transfer_ready(self.timeout)
                        res = True
                    elif self.check_crystal_state(sample_uuid) == "pending_and_current":
                        logging.getLogger("user_level_log").info(
                            "Putting Harvester in Transfer Mode"
                        )
                        self.transfer_sample()
                        if wait_before_load:
                            self._wait_sample_transfer_ready(self.timeout)
                        res = True
                    else:
                        # logging.getLogger("user_level_log").info("ERROR: Sample Could not be Harvested (Harvester Ready, ) ")
                        msg = self.get_status()
                        logging.getLogger("user_level_log").exception(
                            "ERROR: Sample Could not be Harvested"
                        )
                        logging.getLogger("user_level_log").exception(msg)

                        res = False

                    return res
                except RuntimeError:
                    return False

            elif self._ready_to_transfer():
                try:
                    if (
                        self.current_crystal_state(sample_uuid)
                        == "waiting_for_transfer"
                    ):
                        logging.getLogger("user_level_log").info(
                            "Sample Already Harvested, continue"
                        )
                        res = True
                    else:
                        self.abort()
                        self.park()
                        self._wait_ready(self.timeout)
                        logging.getLogger("user_level_log").info("Trash current Sample")
                        self.trash_sample()
                        self._wait_ready(self.timeout)
                        if (
                            self.current_crystal_state(sample_uuid)
                            == "ready_to_execute"
                            or self.current_crystal_state(sample_uuid)
                            == "needs_repositionning"
                        ):
                            logging.getLogger("user_level_log").info(
                                "Harvesting started"
                            )
                            self.harvest_crystal(sample_uuid)
                            if wait_before_load:
                                self._wait_sample_transfer_ready(self.timeout)
                            res = True
                        else:
                            msg = self.get_status()
                            logging.getLogger("user_level_log").info(
                                "Warning: Sample Could not be Harvested Try Again"
                            )
                            return self.harvest_sample_before_mount(sample_uuid)

                    return res
                except RuntimeError:
                    return False
            elif (
                "Harvesting" in self.get_status()
                or self.get_status() == "Finishing Harvesting"
            ):
                logging.getLogger("user_level_log").info(
                    "Warning: Harvesting In Progress Try Again"
                )
                self._wait_sample_transfer_ready(self.timeout)
                return self.harvest_sample_before_mount(sample_uuid)
            else:
                msg = self.get_status()
                logging.getLogger("user_level_log").exception(
                    "ERROR: Sample Could not be Harvested"
                )
                logging.getLogger("user_level_log").exception(msg)
                # Try an abort and move to next sample
                self.abort()
                self.park()
                self._wait_ready(self.timeout)
                return False
        else:
            msg = self.get_status()
            logging.getLogger("user_level_log").exception("ERROR: No sample uuid Found")
            logging.getLogger("user_level_log").exception(msg)
            # Try an abort and move to next sample
            return False

    def get_offsets_for_sample_centering(self) -> tuple[float]:
        """Calculate sample centering offsets
        based on Harvested pin shape pre-calculated offsets

        Return (tuple(float)): (phiy_offset, centringFocus, centringTableVertical)
        """

        pin_to_beam = tuple(self.get_calibrated_pin_offset())

        sample_drift_x = float(self.get_last_sample_drift_offset_x())
        sample_drift_y = float(self.get_last_sample_drift_offset_y())
        sample_drift_z = -float(self.get_last_sample_drift_offset_z())

        pin_cut_shape_x = float(self.get_last_pin_cut_shape_offset_x())
        pin_cut_shape_y = float(self.get_last_pin_cut_shape_offset_y())

        phiy_offset = sample_drift_x - pin_cut_shape_x + float(pin_to_beam[1])

        centringFocus = sample_drift_z + float(pin_to_beam[0])

        centringTableVertical = sample_drift_y - pin_cut_shape_y + float(pin_to_beam[2])

        return (phiy_offset, centringFocus, centringTableVertical)



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging

from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor

__author__ = "Jan Meyer"
__email__ = "jan.meyer@desy.de"
__copyright__ = "(c)2016 DESY, FS-PE, P11"
__license__ = "GPL"


class MotorWPositions(AbstractMotor, Device):
    """
    <object class="MotorWPositions">
        <username>Dummy</username>
        <motors>
            <object role="rolename" href="/dummy"></object>
        </motors>
        <deltas>
            <rolename>0.1</rolename>
        </deltas>
        <positions>
            <position>
                <name>Zoom 1X</name>
                <rolename>0.12</rolename>
            </position>
            <position>
                <name>Zoom 6X</name>
                <rolename>1.23</rolename>
            </position>
        </positions>
    </object>
    """

    def __init__(self, name):
        AbstractMotor.__init__(self, name)
        super().__init__(name)
        self.predefined_positions = {}
        self.motor = None
        self.delta = 0.001
        self.positions = {}
        self._last_position_name = None

    def init(self):
        self._last_position_name = None
        try:
            roles = self["motors"].get_roles()
            role = roles[0]
            self.motor = self.get_object_by_role(role)
        except KeyError:
            self.log.error("MotorWPositions: motor not defined")

            self.log.exception("")
            return
        try:
            self.delta = self["deltas"].get_property(role)
        except Exception:
            logging.getLogger().info(
                "MotorWPositions: no delta defined, setting to %f", self.delta
            )
        try:
            positions = self["positions"]
        except Exception:
            logging.getLogger().error("MotorWPositions: no positions defined")
        else:
            for position in positions:
                name = position.get_property("name")
                pos = position.get_property(role)
                self.predefined_positions[name] = pos
        self.connect(self.motor, "stateChanged", self.motor_state_changed)
        self.connect(self.motor, "valueChanged", self.motor_position_changed)
        self.set_is_ready(self.motor.is_ready())

    def get_limits(self):
        return (1, len(self.predefined_positions))

    def get_predefined_positions_list(self):
        return sorted(self.predefined_positions.keys())
        # return self.predefined_positions

    def get_current_position_name(self, pos=None):
        if pos is None:
            pos = self.motor.get_value()
        for position_name, position in self.predefined_positions.items():
            if self.delta >= abs(pos - position):
                return position_name
        return ""

    def moveToPosition(self, position_name):
        try:
            self.motor.set_value(self.predefined_positions[position_name])
        except Exception:
            self.log.exception("MotorWPositions: invalid position name")

    def setNewPredefinedPosition(self, positionName, positionOffset):
        raise NotImplementedError

    def get_motor_mnemonic(self):
        """
        Descript. :
        """
        return self.actuator_name

    def motor_state_changed(self, state):
        self.updateState(state)
        self.emit("stateChanged", (state,))

    def motor_position_changed(self, absolute_position=None):
        if absolute_position is None:
            absolute_position = self.motor.get_value()
        position_name = self.get_current_position_name(absolute_position)
        if self._last_position_name != position_name:
            self._last_position_name = position_name
            self.emit(
                "predefinedPositionChanged",
                (position_name, position_name and absolute_position or None),
            )
        self.emit("valueChanged", (absolute_position,))

    def updateState(self, state=None):
        """
        Descript. :
        """
        if state is None:
            state = self.get_state()
        self.set_is_ready(state > AbstractMotor.UNUSABLE)

    def get_state(self):
        """
        Descript. : return motor state
        """
        return self.motor.get_state()

    def get_value(self):
        """
        Descript. :
        """
        return self.motor.get_value()

    def set_value(self, absolute_position, timeout=0):
        """
        Descript. :
        """
        self.motor.set_value(absolute_position, timeout)

    def set_value_relative(self, relative_position, timeout=0):
        """
        Descript. :
        """
        self.motor.set_value_relative(relative_position, timeout)

    def stop(self):
        """
        Descript. :
        """
        self.motor.stop()

    def is_moving(self):
        """
        Descript. : True if the motor is currently moving
        """
        return self.motor.is_moving()



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


from mxcubecore.HardwareObjects import QtGraphicsLib as GraphicsLib
from mxcubecore.HardwareObjects.QtGraphicsManager import QtGraphicsManager
from mxcubecore.utils import qt_import

SEQ_ITEM_COLORS = (
    qt_import.QColor(204, 255, 204),
    qt_import.QColor(254, 254, 121),
    qt_import.QColor(72, 191, 255),
    qt_import.QColor(142, 69, 133, 70),
)


__credits__ = ["EMBL Hamburg"]
__category__ = "General"


class EMBLSSXChip(QtGraphicsManager):
    def __init__(self, name):
        QtGraphicsManager.__init__(self, name)

        self.chip_config_list = None
        self.current_chip_index = None

        self.dg_channels_list = []
        self.dg_channel_list_one_zero = []
        self.letters_for_descriptors_list = []
        self.compartment_enable_list = []

        self.channels = []
        self.channels_draw_items = []
        self.pos_coord = [None, None, None]

        self.graphics_coord_axes_item = None

    def init(self):
        """Reads config xml, initiates all necessary hwobj, channels and cmds"""

        self.chip_config_list = eval(self.get_property("chip_properties", "[]"))
        self.current_chip_index = 0

        self.chip_settings = {
            "scan_rate": 0,
            "quarter_density": 0,
            "meandering": 0,
            "old_num_of_exp": 1,
            "num_channels": self.get_property("num_channels"),
            "channels": eval(self.get_property("channels")),
            "default_seq": eval(self.get_property("default_sequence")),
        }

        self.graphics_view = GraphicsLib.GraphicsView()
        self.graphics_view.scene().setSceneRect(0, 0, 340, 140)
        self.graphics_coord_axes_item = GraphicsItemCoordAxes(
            self, self.chip_settings["num_channels"]
        )
        self.graphics_view.graphics_scene.addItem(self.graphics_coord_axes_item)

    def get_chip_config_list(self):
        return self.chip_config_list

    def get_current_chip_config(self):
        return dict(
            self.chip_config_list[self.current_chip_index].items()
            + self.chip_settings.items()
        )

    def get_current_chip_index(self):
        return self.current_chip_index

    def set_current_chip_index(self, index):
        self.current_chip_index = index

    def get_chip_config_str_list(self):
        config_str_list = []
        for chip_config in self.chip_config_list:
            config_str_list.append(
                "Comp: %dx%d, Crystal: %dx%d"
                % (
                    chip_config["num_comp_h"],
                    chip_config["num_comp_v"],
                    chip_config["num_crystal_h"],
                    chip_config["num_crystal_v"],
                )
            )

        return config_str_list

    def set_config_item(self, item_name, item_value):
        if item_name in self.chip_config_list[self.current_chip_index]:
            self.chip_config_list[self.current_chip_index][item_name] = item_value
        else:
            self.log.warning("Item %s not found in the config dict" % item_name)

    def get_dg_channels_list(self):
        return self.dg_channels_list

    def set_dg_channels_list(self, dg_channels_list):
        self.dg_channels_list = dg_channels_list

    def get_dg_channel_list_one_zero(self):
        return self.dg_channel_list_one_zero

    def set_dg_channel_list_one_zero(self, dg_channel_list_one_zero):
        self.dg_channel_list_one_zero = dg_channel_list_one_zero

    def set_channels(self, channels):
        self.chip_settings["channels"] = channels
        self.graphics_coord_axes_item.set_channels(channels)
        self.graphics_view.graphics_scene.update()

    def get_interlacings_list(self):
        current_chip_config = self.chip_config_list[self.current_chip_index]

        # 4 lists of divisors
        crystal_h_divisor_list = None
        crystal_v_divisor_list = None
        comp_h_divisor_list = None
        comp_v_divisor_list = None

        # quarter density check box state 1 - not checked; 2 - checked
        multiplier = 1  # for quarter density usage
        if current_chip_config["quarter_density"] == 1:
            multiplier = 2
        else:
            multiplier = 1

        # get values of the number of crystalalals and number of compartments
        num_crystal_h = current_chip_config["num_crystal_h"] / float(multiplier)
        num_crystal_v = current_chip_config["num_crystal_v"] / float(multiplier)
        num_comp_h = current_chip_config["num_crystal_h"]
        num_comp_v = current_chip_config["num_crystal_v"]

        # set list variables with corresponding lists
        crystal_h_divisor_list = self.get_divisors_list(num_crystal_h)
        crystal_v_divisor_list = self.get_divisors_list(num_crystal_v)
        comp_h_divisor_list = self.get_divisors_list(num_comp_h)
        comp_v_divisor_list = self.get_divisors_list(num_comp_v)

        # remove zero elements from V crystalalal and compartment lists
        del crystal_v_divisor_list[0]
        del comp_v_divisor_list[0]

        # multiply V crystalalal list with number of H crystalals
        for element in range(0, len(crystal_v_divisor_list)):
            crystal_v_divisor_list[element] = (
                crys_v_divisor_list[element] * num_crystal_h
            )

        # multiply V compartment list with number of H compartments
        for element in range(0, len(comp_v_divisor_list)):
            comp_v_divisor_list[element] = comp_v_divisor_list[element] * num_comp_h

        # create a list which contains two lists of H and V (already multiplied)
        # crystalalals
        a_list = sorted(crystal_h_divisor_list + crystal_v_divisor_list)

        # create a list which contains two lists of H and V (already multiplied)
        # compartments
        b_list = comp_h_divisor_list + comp_v_divisor_list
        b_list.sort()
        del b_list[0]  # delete first element

        # multiply B_list with number of H and V crystalalals
        for element in range(0, len(b_list)):
            b_list[element] = b_list[element] * num_crystal_h * num_crystal_v

        # concatenate a_list and b_list the sort the result
        output_list = a_list + b_list
        output_list.sort()

        return output_list

    def get_divisors_list(self, number):
        divisor_list = []

        for x in range(1, number + 1):
            if (number % x) == 0:
                divisor_list.append(x)

        if len(divisor_list) == 0:
            divisor_list.append(0)

        return divisor_list

    # creates and shows the shortlist
    def get_short_list(self):
        text_line = ""

        for element in self.dg_channels_list:
            if inte(lement).checkState() == 2:
                self.dg_channel_list_one_zero.append(1)
            else:
                self.dg_channel_list_one_zero.append(0)

        # line counter
        line_number = 1
        # enabled compartment counter
        count = 0
        # boolean to stop iterating
        done = False

        current_chip_config = self.get_current_chip_config()
        num_comp_v = current_chip_config["num_comp_v"]
        num_comp_h = current_chip_config["num_comp_h"]

        if current_chip_config["quarter_density"] == 1:
            # reduce number of features by 2 if quarter density is enabled
            num_crystal_h = current_chip_config["num_crystal_h"] / 2
            num_crystal_v = current_chip_config["num_crystal_v"] / 2
            line_number = 1

            # loops through all compartments and features
            for n_cv in range(0, num_comp_v):
                for n_ch in range(0, num_comp_h):
                    # checks for enabled compartments
                    if (
                        n_cv != self.compartment_enable_list[count][0]
                        or n_ch != self.compartment_enable_list[count][1]
                    ):
                        continue

                    for n_fv in range(0, num_crystal_v):
                        for n_fh in range(0, num_crystal_h):
                            for exposure in range(
                                0, len(self.dg_channel_list_one_zero) / 4
                            ):
                                # meandering
                                if current_chip_config["meandering"]:
                                    # checks whether n_fv is odd or even, then creates
                                    # modified_n_fh
                                    if n_fv % 2 != 0:
                                        modified_n_fh = num_crystal_h - n_fh - 1
                                    else:
                                        modified_n_fh = n_fh
                                else:
                                    modified_n_fh = n_fh

                                line_keeper = (
                                    str(line_number)
                                    + ";"
                                    + self.from_loop_to_string(
                                        n_cv, n_ch, 2 * n_fv, 2 * modified_n_fh
                                    )
                                    + ";"
                                    + str(0)
                                    + ";"
                                    + str(nCv)
                                    + ";"
                                    + str(nCh)
                                    + ";"
                                    + str(2 * nFv)
                                    + ";"
                                )
                                line_keeper = (
                                    line_keeper
                                    + str(2 * n_fh)
                                    + ";"
                                    + str(0)
                                    + ";"
                                    + str(0)
                                    + ";"
                                    + str(dg_channel_list_one_zero[exposure * 4])
                                    + ";"
                                    + str(
                                        self.dg_channes_list_one_zero[exposure * 4 + 1]
                                    )
                                    + ";"
                                )
                                line_keeper = (
                                    line_keeper
                                    + str(dg_channel_list_one_zero[exposure * 4 + 2])
                                    + ";"
                                    + str(
                                        self.dg_channes_list_one_zero[exposure * 4 + 3]
                                    )
                                    + ";"
                                    + str(0)
                                    + ";\n"
                                )
                                text_line = textLine + line_keeper

                                line_number = line_number + 1  # update line number

                    count = count + 1
                    if count > len(self.compartment_enable_list) - 1:
                        done = True
                        break
                # break the whole loop system when count is bigger than the enabled
                # compartments list length
                if done == True:
                    break

        else:
            line_number = 1
            # loops through all compartments and features
            for n_cv in range(0, num_comp_v):
                for n_ch in range(0, num_comp_h):
                    # checks for enabled compartments
                    if (
                        n_cv != self.compartment_enable_list[count][0]
                        or n_ch != self.compartment_enable_list[count][1]
                    ):
                        continue

                    for n_fv in range(0, num_crystal_v):
                        for n_fh in range(0, num_crystal_h):
                            for exposure in range(0, len(dg_channel_list_one_zero) / 4):
                                # meandering
                                if meandering == 1:
                                    # checks whether n_fv is odd or even, then creates
                                    # modified_n_fh
                                    if n_fv % 2 != 0:
                                        modified_n_fh = num_crystal_h - n_fh - 1
                                    else:
                                        modified_n_fh = n_fh
                                else:
                                    modified_n_fh = n_fh

                                line_keeper = (
                                    str(line_number)
                                    + ";"
                                    + self.from_loop_to_string(
                                        n_cv, n_ch, n_fv, modified_n_fh
                                    )
                                    + ";"
                                    + str(0)
                                    + ";"
                                    + str(nCv)
                                    + ";"
                                    + str(nCh)
                                    + ";"
                                    + str(nFv)
                                    + ";"
                                )
                                line_keeper = (
                                    line_keeper
                                    + str(n_fh)
                                    + ";"
                                    + str(0)
                                    + ";"
                                    + str(0)
                                    + ";"
                                    + str(self.dg_channel_list_one_zero[exposure * 4])
                                    + ";"
                                    + str(
                                        self.dg_channes_list_one_zero[exposure * 4 + 1]
                                    )
                                    + ";"
                                )
                                line_keeper = (
                                    line_keeper
                                    + str(
                                        self.dg_channel_list_one_zero[exposure * 4 + 2]
                                    )
                                    + ";"
                                    + str(
                                        self.dg_channes_list_one_zero[exposure * 4 + 3]
                                    )
                                    + ";"
                                    + str(0)
                                    + ";\n"
                                )
                                text_line = text_line + line_keeper

                                line_number = line_number + 1  # update line number
                    count = count + 1
                    if count > len(self.compartment_enable_list) - 1:
                        done = True
                        break
                # break the whole loop system when count is bigger than the enabled
                # compartments list length
                if done == True:
                    break

        # remove disabled compartments from the string
        # enabledCompartments_text_line = self.remove_disabled_compartments(textLine)
        # add text to the screen

        self.dg_channel_list_one_zero = []

        return text_line

    # converts 4 numbers from loops into a string (B3_cd)
    def from_loop_to_string(self, n_cv, n_ch, n_fv, n_fh):
        capital_letter_start_point = ord("A")
        # smallLetterStartPoint = ord('a')

        # set the int value of character A and a in unicode
        output_string = (
            chr(capital_letter_start_point + n_cv)
            + str(n_ch + 1)
            + "_"
            + self.letters_for_descriptors_list[n_fv]
            + self.letters_for_descriptors_list[n_fh]
        )

        return output_string

    # saves chip data in a new file
    def save_chip_data(self, filename):
        f = open(filename, "w+")
        f.write("[main]\n")

        current_chip_config = self.chip_config_list[self.current_chip_index]
        string_line = "%.9E" % Decimal(current_chip_config["crystal_h_pitch"])
        f.write("chip data.crystal H pitch (um)=%s\n" % string_line.replace(".", ","))
        string_line = "%.9E" % Decimal(current_chip_config["crystal_v_pitch"])
        f.write("chip data.crystal V pitch (um)=%s\n" % string_line.replace(".", ","))
        string_line = "%.9E" % Decimal(current_chip_config["comp_h_pitch"])
        f.write(
            "chip data.compartment H pitch (um)=%s\n" % string_line.replace(".", ",")
        )
        string_line = "%.9E" % Decimal(current_chip_config["comp_v_pitch"])
        f.write(
            "chip data.compartment V pitch (um)=%s\n" % string_line.replace(".", ",")
        )
        f.write("chip data.num crystals H=%d\n" % current_chip_config["num_crystal_h"])
        f.write("chip data.num crystals V=%d\n" % current_chip_config["num_crystal_v"])
        f.write("chip data.num compartment H=%d\n" % current_chip_config["num_comp_h"])
        f.write("chip data.num compartment V=%d\n" % current_chip_config["num_comp_v"])
        f.close()

    # returns a list of letters for descriptors
    def create_descriptors_list(self):
        self.letters_for_descriptors_list = []
        capital_letter_start_point = ord("A")
        small_letter_start_point = ord("a")
        number_start_point = ord("0")

        for element in range(0, 26):
            self.letters_for_descriptors_list.append(
                chr(small_letter_start_point + element)
            )

        for element in range(0, 26):
            self.letters_for_descriptors_list.append(
                chr(capital_letter_start_point + element)
            )

        for element in range(0, 10):
            self.letters_for_descriptors_list.append(chr(number_start_point + element))


class GraphicsItemCoordAxes(GraphicsLib.GraphicsItem):
    def __init__(self, parent, num_channels):
        GraphicsLib.GraphicsItem.__init__(self, parent, position_x=0, position_y=0)

        self.num_channels = num_channels
        self.channels = []

        self.custom_pen.setWidth(1)
        self.custom_pen.setColor(qt_import.Qt.black)

    def set_channels(self, channels):
        self.channels = channels

    def paint(self, painter, option, widget):
        scene_width = self.scene().width()
        scene_height = self.scene().height()

        self.custom_brush.setColor(qt_import.QColor(72, 191, 255))
        painter.setBrush(self.custom_brush)

        offset_x = 5
        offset_y = 0
        height = 30

        self.custom_pen.setColor(qt_import.QColor(qt_importQt.lightGray))
        painter.setPen(self.custom_pen)

        if self.channels:
            self.custom_pen.setColor(qt_import.QColor(qt_importQt.lightGray))
            self.custom_pen.setStyle(qt_import.Qt.DotLine)
            painter.setPen(self.custom_pen)
            # Draw delay reference lines

            max_length = 0
            for channels in self.channels:
                if channels is not None:
                    if channels[1] + channels[2] > max_length:
                        max_length = channels[1] + channels[2]

            self.custom_pen.setColor(qt_import.QColor(Qt.black))
            self.custom_pen.setStyle(qt_import.Qt.SolidLine)
            painter.setPen(self.custom_pen)

            for index, channels in enumerate(self.channels[::-1]):
                if channels is not None:
                    start_x = (
                        offset_x
                        + (scene_width - offset_x * 2) / max_length * channels[1]
                    )
                    size_x = (scene_width - offset_x * 2) / max_length * channels[2]
                    # start_y = scene_height - (scene_height - offset_y - 20) / self.num_channels * index  - offset_y - scene_height / self.num_channels / 2
                    start_y = scene_height - height * (index + 1) - offset_y
                    painter.drawRect(start_x, start_y, size_x, height)

        # Draw x and y axes
        painter.drawLine(offset_x, offset_y, offset_x, scene_height - offset_y)
        for index in range(self.num_channels):
            y_pos = (
                scene_height
                - (scene_height - offset_y - 20) / self.num_channels * index
                - offset_y
            )
            painter.drawLine(offset_x, y_pos, scene_width - 10, y_pos)



from devtools import debug

from mxcubecore.HardwareObjects.ESRF.queue_entry.ssx_base_queue_entry import (
    BaseUserCollectionParameters,
    SsxBaseQueueEntry,
    SsxBaseQueueTaskParameters,
)
from mxcubecore.model.common import LegacyParameters
from mxcubecore.model.queue_model_objects import DataCollection

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class TestUserCollectionParameters(BaseUserCollectionParameters):
    class Config:
        extra: "ignore"


class TestCollectionTaskParameters(SsxBaseQueueTaskParameters):
    user_collection_parameters: TestUserCollectionParameters
    legacy_parameters: LegacyParameters

    @staticmethod
    def update_dependent_fields(field_data):
        new_data = {}
        return new_data


class TestCollectionQueueModel(DataCollection):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class TestCollectionQueueEntry(SsxBaseQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    QMO = TestCollectionQueueModel
    DATA_MODEL = TestCollectionTaskParameters
    NAME = "Test Collection"
    REQUIRES = ["point", "line", "no_shape", "chip", "grid"]

    def __init__(self, view, data_model: TestCollectionQueueModel):
        super().__init__(view=view, data_model=data_model)

    def execute(self):
        super().execute()
        debug(self._data_model._task_data)

        # data_root_path = HWR.beamline.session.get_image_directory(
        #     os.path.join(
        #         self._data_model._task_data.path_parameters.subdir,
        #         self._data_model._task_data.path_parameters.experiment_type,
        #     )
        # )

        # process_path = os.path.join(
        #     HWR.beamline.session.get_base_process_directory(),
        #     self._data_model._task_data.path_parameters.subdir,
        # )

        # subprocess.Popen(
        #     "mkdir --parents %s" % (data_root_path),
        #     shell=True,
        #     stdin=None,
        #     stdout=None,
        #     stderr=None,
        #     close_fds=True,
        # ).wait()

        # dcg = HWR.beamline.lims.icat_client.create_ssx_data_collection_group()
        # HWR.beamline.lims.icat_client.create_ssx_data_collection()

    def pre_execute(self):
        super().pre_execute()

    def post_execute(self):
        super().post_execute()

    def stop(self):
        super().stop()



import glob
import logging
import os
import time
from pathlib import Path
from typing import Tuple

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects import Session
from mxcubecore.model import queue_model_objects


class ESRFSession(Session.Session):
    def __init__(self, name):
        Session.Session.__init__(self, name)

    def init(self):
        Session.Session.init(self)
        self._use_acronym = self.get_property("use_acronym", True)

        archive_base_directory = self["file_info"].get_property(
            "archive_base_directory"
        )
        if archive_base_directory:
            archive_folder = os.path.join(
                self["file_info"].get_property("archive_folder"), time.strftime("%Y")
            )
            queue_model_objects.PathTemplate.set_archive_path(
                archive_base_directory, archive_folder
            )

    def set_endstation_name(self, name: str) -> None:
        name = name.lower().replace("-", "")
        name = {
            "id231": "id23eh1",
            "id232": "id23eh2",
        }.get(name, name)

        self.log.info(f"Setting end-station name to {name}")
        self._endstation_name = name

    def get_full_paths(self, subdir: str, tag: str) -> Tuple[str, str]:
        """
        Returns the full path to both image and processed data.
        The path(s) returned will follow the convention:

          <base_direcotry>/<subdir>/run_<NUMBER>_<tag>

        Where NUMBER is a automatically sequential number and
        base_directory the path returned by get_base_image/process_direcotry

        :param subdir: subdirecotry
        :param tag: tag for

        :returns: Tuple with the full path to image and processed data
        """
        subdir = Path(self._sanitize_subdir(subdir))

        if not subdir.is_relative_to(self.get_base_image_directory()):
            error_message = "Invalid subdirectory"
            logging.getLogger("user_level_log").error(error_message)
            raise PermissionError(error_message)

        folders = glob.glob(
            os.path.join(self.get_base_image_directory(), subdir) + "/run*"
        )

        runs = [0]
        for folder in folders:
            runs.append(int(folder.split("/")[-1].strip("run_").split("_")[0]))

        run_num = max(runs) + 1

        # Use the same sequence numbering for all tags/types, (We are not
        # creating individual run number per tag)
        full_path = os.path.join(
            self.get_base_image_directory(), subdir, f"run_{run_num:02d}_{tag}/"
        )

        # Check collects in queue not yet collected
        for pt in HWR.beamline.queue_model.get_path_templates():
            if pt[1].directory.startswith(full_path[:-1]):
                run_num += 1

                full_path = os.path.join(
                    self.get_base_image_directory(), subdir, f"run_{run_num:02d}/"
                )

        full_path = os.path.join(
            self.get_base_image_directory(), subdir, f"run_{run_num:02d}_{tag}/"
        )

        if not subdir.is_relative_to(self.get_base_process_directory()):
            error_message = "Invalid subdirectory"
            logging.getLogger("user_level_log").error(error_message)
            raise PermissionError(error_message)

        process_path = os.path.join(
            self.get_base_process_directory(), subdir, f"run_{run_num:02d}_{tag}/"
        )

        return full_path, process_path

    def get_default_subdir(self, sample_data: dict) -> str:
        """
        Gets the default sub-directory based on sample information

        Args:
           sample_data: Lims sample dictionary

        Returns:
           Sub-directory path string
        """
        subdir = ""

        if isinstance(sample_data, dict):
            sample_name = sample_data.get("sampleName", "")
            protein_acronym = sample_data.get("proteinAcronym", "")
        else:
            sample_name = sample_data.name
            protein_acronym = sample_data.crystals[0].protein_acronym

        if self._use_acronym and protein_acronym:
            subdir = "%s/%s-%s/" % (protein_acronym, protein_acronym, sample_name)
        else:
            subdir = "%s/" % sample_name

        return subdir.replace(":", "-")

    def get_image_directory(self, sub_dir=None):
        """
        Returns the full path to images, using the name of each of
        data_nodes parents as sub directories.

        :param data_node: The data node to get additional
                          information from, (which will be added
                          to the path).
        :type data_node: TaskNode

        :returns: The full path to images.
        :rtype: str
        """
        data_path = self.get_base_image_directory()

        if sub_dir:
            run_num = 0
            data_path = os.path.join(
                self.get_base_image_directory(), sub_dir, f"run_{run_num}/"
            )

            while os.path.exists(data_path):
                data_path = os.path.join(
                    self.get_base_image_directory(), sub_dir, f"run_{run_num}/"
                )
                run_num += 1

        return data_path



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import copy
import datetime
import logging
import os
import pickle
import time
from math import sqrt

import beam_align
import gevent
import h5py
import numpy as np
import optical_alignment
import scan_and_align
from anneal import anneal as anneal_procedure
from camera import camera
from detector import detector
from goniometer import goniometer
from scipy.optimize import minimize

from mxcubecore.model.queue_model_enumerables import CENTRING_METHOD

try:
    import lmfit
    from lmfit import fit_report
except ImportError:
    logging.warning(
        "Could not lmfit minimization library, "
        + "refractive model centring will not work."
    )

try:
    import lucid2 as lucid
except ImportError:
    try:
        import lucid
    except ImportError:
        logging.warning(
            "Could not find autocentring library, automatic centring is disabled"
        )

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.GenericDiffractometer import GenericDiffractometer
from mxcubecore.TaskUtils import task

__credits__ = ["SOLEIL"]
__version__ = "2.3."
__category__ = "General"


class PX2Diffractometer(GenericDiffractometer):
    """
    Description:
    """

    AUTOMATIC_CENTRING_IMAGES = 6

    motor_name_mapping = [
        ("AlignmentX", "phix"),
        ("AlignmentY", "phiy"),
        ("AlignmentZ", "phiz"),
        ("CentringX", "sampx"),
        ("CentringY", "sampy"),
        ("Omega", "phi"),
        ("Kappa", "kappa"),
        ("Phi", "kappa_phi"),
        ("beam_x", "beam_x"),
        ("beam_y", "beam_y"),
    ]

    def __init__(self, *args):
        """
        Description:
        """
        GenericDiffractometer.__init__(self, *args)

        # Hardware objects ----------------------------------------------------
        self.zoom_motor_hwobj = None
        self.omega_reference_motor = None
        self.centring_hwobj = None
        self.minikappa_correction_hwobj = None
        self.nclicks = None
        self.step = None
        self.centring_method = None
        self.collecting = False

        # Channels and commands -----------------------------------------------
        self.chan_calib_x = None
        self.chan_calib_y = None
        self.chan_current_phase = None
        self.chan_head_type = None
        self.chan_fast_shutter_is_open = None
        self.chan_state = None
        self.chan_status = None
        self.chan_sync_move_motors = None
        self.chan_scintillator_position = None
        self.chan_capillary_position = None
        self.cmd_start_set_phase = None
        self.cmd_start_auto_focus = None
        self.cmd_get_omega_scan_limits = None
        self.cmd_save_centring_positions = None
        self.centring_time = None
        # Internal values -----------------------------------------------------
        self.use_sc = False
        self.omega_reference_pos = [0, 0]
        self.reference_pos = [680, 512]

        self.goniometer = goniometer()
        self.camera = camera()
        self.detector = detector()

        self.md2_to_mxcube = dict(
            [(key, value) for key, value in self.motor_name_mapping]
        )
        self.mxcube_to_md2 = dict(
            [(value, key) for key, value in self.motor_name_mapping]
        )

        self.log = self.log

    def init(self):
        """
        Description:
        """
        GenericDiffractometer.init(self)
        self.centring_status = {"valid": False}

        self.chan_state = self.get_channel_object("State")
        self.chan_status = self.get_channel_object("Status")

        self.current_state = self.chan_state.get_value()
        self.current_status = self.chan_status.get_value()

        self.chan_state.connect_signal("update", self.state_changed)
        self.chan_status.connect_signal("update", self.status_changed)

        self.chan_calib_x = self.get_channel_object("CoaxCamScaleX")
        self.chan_calib_y = self.get_channel_object("CoaxCamScaleY")
        self.update_pixels_per_mm()

        self.chan_head_type = self.get_channel_object("HeadType")
        self.head_type = self.chan_head_type.get_value()

        self.chan_current_phase = self.get_channel_object("CurrentPhase")
        self.connect(self.chan_current_phase, "update", self.current_phase_changed)

        self.chan_fast_shutter_is_open = self.get_channel_object("FastShutterIsOpen")
        self.chan_fast_shutter_is_open.connect_signal(
            "update", self.fast_shutter_state_changed
        )

        self.chan_scintillator_position = self.get_channel_object(
            "ScintillatorPosition"
        )
        self.chan_capillary_position = self.get_channel_object("CapillaryPosition")

        self.cmd_start_set_phase = self.get_command_object("startSetPhase")
        self.cmd_start_auto_focus = self.get_command_object("startAutoFocus")
        self.cmd_get_omega_scan_limits = self.get_command_object(
            "getOmegaMotorDynamicScanLimits"
        )
        self.cmd_save_centring_positions = self.get_command_object(
            "saveCentringPositions"
        )

        self.centring_hwobj = self.get_object_by_role("centring")
        self.minikappa_correction_hwobj = self.get_object_by_role(
            "minikappa_correction"
        )

        self.zoom_motor_hwobj = self.get_object_by_role("zoom")
        self.connect(self.zoom_motor_hwobj, "valueChanged", self.zoom_position_changed)
        self.connect(
            self.zoom_motor_hwobj,
            "predefinedPositionChanged",
            self.zoom_motor_predefined_position_changed,
        )

        self.connect(self.motor_hwobj_dict["phi"], "valueChanged", self.phi_motor_moved)
        self.connect(
            self.motor_hwobj_dict["phiy"], "valueChanged", self.phiy_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["phiz"], "valueChanged", self.phiz_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["kappa"], "valueChanged", self.kappa_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["kappa_phi"],
            "valueChanged",
            self.kappa_phi_motor_moved,
        )
        self.connect(
            self.motor_hwobj_dict["sampx"], "valueChanged", self.sampx_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["sampy"], "valueChanged", self.sampy_motor_moved
        )

        self.omega_reference_par = eval(self.get_property("omega_reference"))
        self.omega_reference_motor = self.get_object_by_role(
            self.omega_reference_par["motor_name"]
        )

        self.connect(
            self.omega_reference_motor,
            "valueChanged",
            self.omega_reference_motor_moved,
        )

        self.omega_reference_motor_moved(self.omega_reference_motor.get_value())

        # self.use_sc = self.get_property("use_sample_changer")

    def use_sample_changer(self):
        """
        Description:
        """
        return not self.in_plate_mode

    def beam_position_changed(self, value):
        self.beam_position = value

    def state_changed(self, state):
        # self.log.debug("State changed: %s" % str(state))
        if self.current_state != state:
            self.current_state = state
            self.emit("minidiffStateChanged", (self.current_state))

    def status_changed(self, status):
        if self.current_status != status:
            self.current_status = status
            self.emit("minidiffStatusChanged", (self.current_status))

    def zoom_position_changed(self, value):
        self.update_pixels_per_mm()
        self.current_motor_positions["zoom"] = value
        self.refresh_omega_reference_position()

    def zoom_motor_predefined_position_changed(self, position_name, offset):
        self.update_pixels_per_mm()
        self.emit("zoomMotorPredefinedPositionChanged", (position_name, offset))

    def omega_reference_add_constraint(self):
        """
        Descript. :
        """
        if self.omega_reference_par is None or self.beam_position is None:
            return
        if self.omega_reference_par["camera_axis"].lower() == "x":
            on_beam = (
                self.beam_position[0] - self.zoom_centre["x"]
            ) * self.omega_reference_par[
                "direction"
            ] / self.pixels_per_mm_x + self.omega_reference_par["position"]
        else:
            on_beam = (
                self.beam_position[1] - self.zoom_centre["y"]
            ) * self.omega_reference_par[
                "direction"
            ] / self.pixels_per_mm_y + self.omega_reference_par["position"]
        self.centring_hwobj.appendMotorConstraint(self.omega_reference_motor, on_beam)

    def omega_reference_motor_moved(self, pos):
        """
        Descript. :
        """
        if self.omega_reference_par["camera_axis"].lower() == "x":
            pos = (
                self.omega_reference_par["direction"]
                * (pos - self.omega_reference_par["position"])
                * self.pixels_per_mm_x
                + self.zoom_centre["x"]
            )
            self.reference_pos = (pos, -10)
        else:
            pos = (
                self.omega_reference_par["direction"]
                * (pos - self.omega_reference_par["position"])
                * self.pixels_per_mm_y
                + self.zoom_centre["y"]
            )
            self.reference_pos = (-10, pos)
        self.emit("omegaReferenceChanged", (self.reference_pos,))

    def fast_shutter_state_changed(self, is_open):
        """
        Description:
        """
        self.fast_shutter_is_open = is_open
        if is_open:
            msg = "Opened"
        else:
            msg = "Closed"
        self.emit("minidiffShutterStateChanged", (self.fast_shutter_is_open, msg))

    def current_phase_changed(self, current_phase):
        """
        Descript. :
        """
        if self.current_phase == current_phase:
            return
        self.current_phase = current_phase
        if current_phase != GenericDiffractometer.PHASE_UNKNOWN:
            logging.getLogger("GUI").info(
                "Diffractometer: Current phase changed to %s" % current_phase
            )
        self.emit("minidiffPhaseChanged", (current_phase,))

    def phi_motor_moved(self, pos):
        """
        Descript. :
        """
        self.current_motor_positions["phi"] = pos
        self.emit("phiMotorMoved", pos)

    def phiy_motor_moved(self, pos):
        self.current_motor_positions["phiy"] = pos

    def phiz_motor_moved(self, pos):
        self.current_motor_positions["phiz"] = pos

    def sampx_motor_moved(self, pos):
        self.current_motor_positions["sampx"] = pos

    def sampy_motor_moved(self, pos):
        self.current_motor_positions["sampy"] = pos

    def kappa_motor_moved(self, pos):
        """
        Descript. :
        """
        self.current_motor_positions["kappa"] = pos
        if time.time() - self.centring_time > 1.0:
            self.invalidate_centring()
        self.emit_diffractometer_moved()
        self.emit("kappaMotorMoved", pos)

    def kappa_phi_motor_moved(self, pos):
        """
        Descript. :
        """
        self.current_motor_positions["kappa_phi"] = pos
        if time.time() - self.centring_time > 1.0:
            self.invalidate_centring()
        self.emit_diffractometer_moved()
        self.emit("kappaPhiMotorMoved", pos)

    def refresh_omega_reference_position(self):
        """
        Descript. :
        """
        if self.omega_reference_motor is not None:
            reference_pos = self.omega_reference_motor.get_value()
            self.omega_reference_motor_moved(reference_pos)

    def update_pixels_per_mm(self, *args):
        """
        Descript. :
        """
        if self.chan_calib_x:
            self.pixels_per_mm_x = 1.0 / self.chan_calib_x.get_value()
            self.pixels_per_mm_y = 1.0 / self.chan_calib_y.get_value()
            self.emit(
                "pixelsPerMmChanged", ((self.pixels_per_mm_x, self.pixels_per_mm_y),)
            )

    def set_phase(self, phase, timeout=60):
        """Sets diffractometer to the selected phase.
        In the plate mode before going to or away from
        Transfer or Beam location phase if needed then detector
        is moved to the safe distance to avoid collision.
        """
        # self.wait_device_ready(2)
        logging.getLogger("GUI").warning(
            "Diffractometer: Setting %s phase. Please wait..." % phase
        )

        if phase in (
            GenericDiffractometer.PHASE_TRANSFER,
            GenericDiffractometer.PHASE_BEAM,
        ) or self.current_phase in (
            GenericDiffractometer.PHASE_TRANSFER,
            GenericDiffractometer.PHASE_BEAM,
        ):
            detector_distance = HWR.beamline.detector.distance.get_value()
            self.log.debug(
                "Diffractometer current phase: %s " % self.current_phase
                + "selected phase: %s " % phase
                + "detector distance: %d mm" % detector_distance
            )
            if detector_distance < 350:
                logging.getLogger("GUI").info("Moving detector to safe distance")
                HWR.beamline.detector.distance.set_value(350)
                self.detector.insert_protective_cover()

        if timeout is not None:
            self.cmd_start_set_phase(phase)
            gevent.sleep(1)
            with gevent.Timeout(
                timeout, Exception("Timeout waiting for phase %s" % phase)
            ):
                while phase != self.chan_current_phase.get_value():
                    gevent.sleep(0.01)
        else:
            self.cmd_start_set_phase(phase)

    def start_auto_focus(self, timeout=None):
        """
        Descript. :
        """
        if timeout:
            self.ready_event.clear()
            set_phase_task = gevent.spawn(
                self.execute_server_task, self.cmd_start_auto_focus(), timeout
            )
            self.ready_event.wait()
            self.ready_event.clear()
        else:
            self.cmd_start_auto_focus()

    def centring_done(self, centring_procedure):
        """
        Descript. :
        """
        try:
            motor_pos = centring_procedure.get()
            # if isinstance(motor_pos, gevent.GreenletExit):
            # raise motor_pos
            # self.log.info('motor_pos from centring_procedure %s' % motor_pos)
        except Exception:
            logging.exception("Could not complete centring")
            self.emit_centring_failed()
        else:
            self.emit_progress_message("Moving sample to centred position...")
            self.emit_centring_moving()
            try:
                self.move_to_motors_positions(motor_pos)
            except Exception:
                logging.exception("Could not move to centred position")
                self.emit_centring_failed()

            if (
                self.current_centring_method
                == GenericDiffractometer.CENTRING_METHOD_AUTO
            ):
                self.emit("newAutomaticCentringPoint", motor_pos)

            self.ready_event.set()
            self.centring_time = time.time()
            self.emit_centring_successful()

    def move_motors(self, motor_positions, timeout=15):
        """
        Moves diffractometer motors to the requested positions

        :param motors_dict:  with motor names or hwobj
                            and target values.
        :type motors_dict: dict
        """

        position = self.translate_from_mxcube_to_md2(motor_positions)

        self.goniometer.set_position(position, timeout=timeout)

    def get_centred_point_from_coord(self, x, y, return_by_names=None):
        """
        Descript. :
        """

        self.log.info("get_centred_point_from_coord: x, y: %s %s" % (x, y))
        self.log.info(
            "get_centred_point_from_coord: self.pixels_per_mm_x, self.pixels_per_mm_ y: %s %s"
            % (self.pixels_per_mm_x, self.pixels_per_mm_y)
        )
        self.log.info(
            "get_centred_point_from_coord: self.beam_position: %s"
            % str(self.beam_position)
        )

        current_position = self.goniometer.get_aligned_position()

        omega_reference = self.omega_reference_par["position"]

        alignmentz_shift = current_position["AlignmentZ"] - omega_reference

        vertical_shift = y - self.beam_position[1]

        horizontal_shift = x - self.beam_position[0]

        vertical_shift /= self.pixels_per_mm_y
        horizontal_shift /= self.pixels_per_mm_x

        self.log.info(
            "get_centred_point_from_coord: original_vertical_shift: %s "
            % vertical_shift
        )

        vertical_shift += alignmentz_shift

        centringx_shift, centringy_shift = self.goniometer.get_x_and_y(
            0, vertical_shift, current_position["Omega"]
        )

        centred_point = copy.deepcopy(current_position)

        self.log.info(
            "get_centred_point_from_coord: alignmentz_shift: %s " % alignmentz_shift
        )
        self.log.info(
            "get_centred_point_from_coord: horizontal_shift: %s " % horizontal_shift
        )

        self.log.info(
            "get_centred_point_from_coord: vertical_shift: %s " % vertical_shift
        )
        self.log.info(
            "get_centred_point_from_coord: centringx_shift: %s " % centringx_shift
        )
        self.log.info(
            "get_centred_point_from_coord: centringy_shift: %s " % centringy_shift
        )

        centred_point["AlignmentZ"] -= alignmentz_shift
        centred_point["AlignmentY"] -= horizontal_shift

        centred_point["CentringX"] += centringx_shift
        centred_point["CentringY"] += centringy_shift

        # centred_point['Omega'] += 90.
        pos = self.translate_from_md2_to_mxcube(centred_point)

        self.log.info("get_centred_point_from_coord: centred_point: %s " % str(pos))

        return pos

        self.centring_hwobj.initCentringProcedure()
        self.centring_hwobj.appendCentringDataPoint(
            {
                "X": (x - self.beam_position[0]) / self.pixels_per_mm_x,
                "Y": (y - self.beam_position[1]) / self.pixels_per_mm_y,
            }
        )
        self.omega_reference_add_constraint()
        pos = self.centring_hwobj.centeredPosition()
        self.log.info("get_centred_point_from_coord: pos %s" % str(pos))
        if return_by_names:
            pos = self.convert_from_obj_to_name(pos)
        return pos

    def move_to_beam(self, x, y, omega=None):
        """Creates a new centring point based on all motors positions"""
        if self.current_phase != "BeamLocation":
            GenericDiffractometer.move_to_beam(
                self, x, y, omega=self.goniometer.get_omega_position()
            )
        else:
            self.log.debug(
                "Diffractometer: Move to screen"
                + " position disabled in BeamLocation phase."
            )

    def manual_centring(
        self,
        n_clicks=3,
        alignmenty_direction=-1.0,
        alignmentz_direction=1.0,
        centringx_direction=-1.0,
        centringy_direction=1.0,
        refractive_model=False,
    ):
        """
        Descript. :
        """
        logging.getLogger("user_level_log").info("starting manual centring")
        _start = time.time()
        result_position = {}
        self.goniometer.insert_backlight()
        self.goniometer.extract_frontlight()

        reference_position = self.goniometer.get_aligned_position()

        vertical_clicks = []
        horizontal_clicks = []
        vertical_discplacements = []
        horizontal_displacements = []
        omegas = []
        images = []
        auxiliary_images = []
        calibrations = []

        if isinstance(self.nclicks, int) and self.nclicks >= 3:
            n_clicks = self.nclicks

        logging.getLogger("user_level_log").info(
            "expected number of clicks %d" % (n_clicks)
        )

        if self.step is not None:
            step = self.step
        else:
            step = 360.0 / (n_clicks)

        logging.getLogger("user_level_log").info("default centring step %.2f" % (step))

        start_clicks = time.time()

        for k in range(n_clicks):
            self.user_clicked_event = gevent.event.AsyncResult()
            x, y = self.user_clicked_event.get()
            image = HWR.beamline.sample_view.camera.get_last_image()
            calibration = self.camera.get_calibration()
            omega = self.goniometer.get_omega_position()

            vertical_clicks.append(y)
            horizontal_clicks.append(x)
            omegas.append(omega)
            images.append(image)
            calibrations.append([calibration])

            x -= self.beam_position[0]
            x /= self.pixels_per_mm_x
            y -= self.beam_position[1]
            y /= self.pixels_per_mm_y
            vertical_discplacements.append(y)
            horizontal_displacements.append(x)

            self.log.info("click %d %f %f %f" % (k + 1, omega, x, y))

            if k <= n_clicks:
                self.goniometer.set_position({"Omega": omega + step})

        end_clicks = time.time()

        name_pattern = "%s_%s" % (os.getuid(), time.asctime().replace(" ", "_"))
        directory = "%s/manual_optical_alignment" % os.getenv("HOME")

        save_history_command = "history_saver.py -s %.2f -e %.2f -d %s -n %s &" % (
            start_clicks,
            end_clicks,
            directory,
            name_pattern,
        )

        self.log.info("save_history_command: %s" % save_history_command)
        os.system(save_history_command)

        vertical_discplacements = np.array(vertical_discplacements) * 1.0e3

        angles = np.radians(omegas)

        if self.centring_method != CENTRING_METHOD.REFRACTIVE:
            initial_parameters = [4.0, 25.0, 0.05]
            fit_y = minimize(
                self.circle_model_residual,
                initial_parameters,
                method="nelder-mead",
                args=(angles, vertical_discplacements),
            )

            c, r, alpha = fit_y.x
            c *= 1e-3
            r *= 1.0e-3
            v = {"c": c, "r": r, "alpha": alpha}

        else:
            initial_parameters = lmfit.Parameters()
            initial_parameters.add_many(
                ("c", 0.0, True, -5e3, +5e3, None, None),
                ("r", 0.0, True, 0.0, 4e3, None, None),
                ("alpha", -np.pi / 3, True, -2 * np.pi, 2 * np.pi, None, None),
                ("front", 0.01, True, 0.0, 1.0, None, None),
                ("back", 0.005, True, 0.0, 1.0, None, None),
                ("n", 1.31, True, 1.29, 1.33, None, None),
                ("beta", 0.0, True, -2 * np.pi, +2 * np.pi, None, None),
            )

            fit_y = lmfit.minimize(
                self.refractive_model_residual,
                initial_parameters,
                method="nelder",
                args=(angles, vertical_discplacements),
            )
            self.log.info(fit_report(fit_y))
            optimal_params = fit_y.params
            v = optimal_params.valuesdict()
            c = v["c"]
            r = v["r"]
            alpha = v["alpha"]
            front = v["front"]
            back = v["back"]
            n = v["n"]
            beta = v["beta"]

            c *= 1.0e-3
            r *= 1.0e-3
            front *= 1.0e-3
            back *= 1.0e-3

        horizontal_center = np.mean(horizontal_displacements)

        d_sampx = centringx_direction * r * np.sin(alpha)
        d_sampy = centringy_direction * r * np.cos(alpha)
        d_y = alignmenty_direction * horizontal_center
        d_z = alignmentz_direction * c

        move_vector_dictionary = {
            "AlignmentZ": d_z,
            "AlignmentY": d_y,
            "CentringX": d_sampx,
            "CentringY": d_sampy,
        }

        for motor in reference_position:
            result_position[motor] = reference_position[motor]
            if motor in move_vector_dictionary:
                result_position[motor] += move_vector_dictionary[motor]

        _end = time.time()
        duration = _end - _start
        self.log.info(
            "input and analysis in manual_centring took %.3f seconds" % duration
        )

        results = {
            "vertical_clicks": vertical_clicks,
            "horizontal_clicks": horizontal_clicks,
            "vertical_discplacements": vertical_discplacements,
            "horizontal_displacements": horizontal_displacements,
            "omegas": omegas,
            "angles": angles,
            "calibrations": calibrations,
            "reference_position": reference_position,
            "result_position": result_position,
            "duration": duration,
            "vertical_optimal_parameters": v,
        }

        template = os.path.join(directory, name_pattern)

        if not os.path.isdir(directory):
            os.makedirs(directory)

        images_filename = "%s_images.h5" % template
        images_file = h5py.File(images_filename, "w")
        images_file.create_dataset(
            "images", data=np.array(images), compression="lzf", dtype=np.uint8
        )
        images_file.close()

        clicks_filename = "%s_clicks.pickle" % template
        f = open(clicks_filename, "w")
        pickle.dump(results, f)
        f.close()

        self.log.info(
            "manual_centring finished in %.3f seconds" % (time.time() - _start)
        )

        translated_position = self.translate_from_md2_to_mxcube(result_position)
        return translated_position

    def translate_from_md2_to_mxcube(self, position):
        translated_position = {}

        for key in position:
            translated_position[self.md2_to_mxcube[key]] = position[key]

        return translated_position

    def convert_from_obj_to_name(self, motor_pos):
        """ """
        # self.log.info('convert_from_obj_to_name motor_pos %s' % str(motor_pos))
        # self.log.info('convert_from_obj_to_name cml %s ' % str(self.centring_motors_list))
        # self.log.info('convert_from_obj_to_name motor_names %s' % str([motor.motor_name for motor in motor_pos]))

        aligned_position = self.goniometer.get_aligned_position()

        motors = self.translate_from_mxcube_to_md2(motor_pos)

        for key in aligned_position:
            if key not in motors:
                motors[key] = aligned_position[key]

        motors = self.translate_from_md2_to_mxcube(motors)
        # motors = {}
        # for motor_role in self.centring_motors_list:
        # self.log.info('motor_role %s' % motor_role)
        # motor_obj = self.get_object_by_role(motor_role)
        # try:
        # motors[motor_role] = motor_pos[motor_obj]
        # except KeyError:

        # if motor_obj:
        # motors[motor_role] = motor_obj.get_value()

        motors["beam_x"] = (
            self.beam_position[0] - self.zoom_centre["x"]
        ) / self.pixels_per_mm_y
        motors["beam_y"] = (
            self.beam_position[1] - self.zoom_centre["y"]
        ) / self.pixels_per_mm_x
        self.log.info("convert_from_obj_to_name motors %s" % str(motors))
        return motors

    def translate_from_mxcube_to_md2(self, position):
        translated_position = {}

        for key in position:
            if isinstance(key, str):
                translated_position[self.mxcube_to_md2[key]] = position[key]
            else:
                translated_position[key.motor_name] = position[key]
        return translated_position

    def circle_model(self, angles, c, r, alpha):
        return c + r * np.cos(angles - alpha)

    def circle_model_residual(self, parameters, angles, data):
        c, r, alpha = parameters
        model = self.circle_model(angles, c, r, alpha)
        # return 1./(2*len(model)) * np.sum(np.sum(np.abs(data - model)**2))
        return self.cost(data, model, normalize=True)

    def circle_model_residual2(self, parameters, angles, data):
        v = parameters.valuesdict()
        c = v["c"]
        r = v["r"]
        alpha = v["alpha"]
        model = self.circle_model(angles, c, r, alpha)

        return self.cost_array(data, model)

    def refractive_model(self, t, c, r, alpha, front, back, n, beta):
        return self.circle_model(t, c, r, alpha) - self.shift(t, front, back, n, beta)

    def refractive_model_residual(self, parameters, angles, data):
        v = parameters.valuesdict()
        c = v["c"]
        r = v["r"]
        alpha = v["alpha"]
        front = v["front"]
        back = v["back"]
        n = v["n"]
        beta = v["beta"]
        model = self.refractive_model(angles, c, r, alpha, front, back, n, beta)

        return self.cost_array(data, model)

    def cost(self, data, model, factor=1.0, normalize=False):
        if normalize == True:
            factor = 1.0 / (2 * len(model))
        return factor * np.sum(np.sum(np.abs(data - model) ** 2))

    def cost_array(self, data, model):
        return np.abs(data - model) ** 2

    def i(self, t, n):
        return np.arcsin(np.sin(t) / n)

    def planparallel_shift(self, depth, t, n, sense=1):
        i = self.i(t, n)
        return -depth * np.sin(sense * t - i) / np.cos(i)

    def shift(self, t, f, b, n, beta):
        t = t - beta
        dt = np.degrees(t)
        s = np.zeros(dt.shape)
        t_base = t % (2 * np.pi)
        mask = np.where(((t_base < 3 * np.pi / 2) & (t_base >= np.pi / 2)), 1, 0)
        s[mask == 0] = self.planparallel_shift(f, t_base[mask == 0], n, sense=1)
        s[mask == 1] = self.planparallel_shift(b, t_base[mask == 1], n, sense=-1)
        return s

    def motor_positions_to_screen(self, centred_positions_dict):
        """
        Descript. :
        """
        c = centred_positions_dict

        # self.log.info('motor_positions_to_screen c %s ' % str(c))

        # kappa = self.current_motor_positions["kappa"]
        # phi = self.current_motor_positions["kappa_phi"]
        self.log.info("centred_positions_dict %s" % str(centred_positions_dict))
        try:
            for key in c:
                if c[key] is None:
                    try:
                        c[key] = self.motor_hwobj_dict[key].get_value()
                    except Exception:
                        # self.log.info('motor_positions_to_screen exception key %s' % key)
                        self.log.exception("")

            if "kappa" in c and c["kappa"] is None:
                kappa = self.motor_hwobj_dict["kappa"].get_value()
                c["kappa"] = kappa
            else:
                c["kappa"] = self.goniometer.get_kappa_position()

            if "kappa_phi" in c and c["kappa_phi"] is None:
                phi = self.motor_hwobj_dict["kappa_phi"].get_value()
                c["kappa_phi"] = phi
            else:
                c["kappa_phi"] = self.goniometer.get_phi_position()

            if "beam_x" in c and c["beam_x"] in [0.0, None]:
                c["beam_x"] = 0.0  # self.beam_position[0]
            else:
                c["beam_x"] = 0.0  # self.beam_position[0]

            if "beam_y" not in c and c["beam_y"] in [0.0, None]:
                c["beam_y"] = 0.0  # self.beam_position[1]
            else:
                c["beam_y"] = 0.0  # self.beam_position[1]

            # self.log.info('motor_positions_to_screen c2 %s ' % str(c))

            if (c["kappa"], c["kappa_phi"]) != (
                self.goniometer.get_kappa_position(),
                self.goniometer.get_phi_position(),
            ) and self.minikappa_correction_hwobj is not None:
                self.log.info("calculating minikappa correction")
                (
                    c["sampx"],
                    c["sampy"],
                    c["phiy"],
                ) = self.minikappa_correction_hwobj.shift(
                    c["kappa"],
                    c["kappa_phi"],
                    [c["sampx"], c["sampy"], c["phiy"]],
                    c["kappa"],
                    c["kappa_phi"],
                )

            xy = self.centring_hwobj.centringToScreen(c)
            # self.log.info('xy %s' % xy)

            if xy:
                x = (xy["X"] + c["beam_x"]) * self.pixels_per_mm_x + self.zoom_centre[
                    "x"
                ]
                y = (xy["Y"] + c["beam_y"]) * self.pixels_per_mm_y + self.zoom_centre[
                    "y"
                ]
                return x, y
        except Exception:
            return 0, 0

    def move_to_centred_position(self, centred_position):
        """
        Descript. :
        """
        # self.log.info('in move_to_centred_position')
        # self.log.info('centred_position %s' % centred_position)

        if centred_position.beam_x is None:
            centred_position.beam_x = 0.0

        if centred_position.beam_y is None:
            centred_position.beam_y = 0.0

        if self.current_phase != "BeamLocation":
            try:
                x, y = centred_position.beam_x, centred_position.beam_y
                dx = (
                    self.beam_position[0] - self.zoom_centre["x"]
                ) / self.pixels_per_mm_x - x
                dy = (
                    self.beam_position[1] - self.zoom_centre["y"]
                ) / self.pixels_per_mm_y - y
                motor_pos = {
                    self.motor_hwobj_dict["sampx"]: centred_position.sampx,
                    self.motor_hwobj_dict["sampy"]: centred_position.sampy,
                    self.motor_hwobj_dict["phi"]: centred_position.phi,
                    self.motor_hwobj_dict["phiy"]: (
                        centred_position.phiy
                        + self.centring_hwobj.camera2alignmentMotor(
                            self.motor_hwobj_dict["phiy"], {"X": dx, "Y": dy}
                        )
                    ),
                    self.motor_hwobj_dict["phiz"]: (
                        centred_position.phiz
                        + self.centring_hwobj.camera2alignmentMotor(
                            self.motor_hwobj_dict["phiz"], {"X": dx, "Y": dy}
                        )
                    ),
                    self.motor_hwobj_dict["kappa"]: centred_position.kappa,
                    self.motor_hwobj_dict["kappa_phi"]: centred_position.kappa_phi,
                }
                self.move_to_motors_positions(motor_pos)
            except Exception:
                logging.exception("Could not move to centred position")
        else:
            self.log.debug("Move to centred position disabled in BeamLocation phase.")

    def move_to_motors_positions(self, motors_positions, wait=False):
        """ """
        self.emit_progress_message("Moving to motors positions...")
        self.move_to_motors_positions_procedure = gevent.spawn(
            self.move_motors, motors_positions
        )
        self.move_to_motors_positions_procedure.link(self.move_motors_done)

    def move_kappa_and_phi(self, kappa=None, kappa_phi=None, wait=False):
        """
        Descript. :
        """
        try:
            return self.move_kappa_and_phi_procedure(kappa, kappa_phi, wait=wait)
        except Exception:
            logging.exception("Could not move kappa and kappa_phi")

    @task
    def move_kappa_and_phi_procedure(self, new_kappa=None, new_kappa_phi=None):
        """
        Descript. :
        """
        kappa = self.motor_hwobj_dict["kappa"].get_value()
        kappa_phi = self.motor_hwobj_dict["kappa_phi"].get_value()

        if new_kappa is None:
            new_kappa = kappa
        if new_kappa_phi is None:
            new_kappa_phi = kappa_phi

        motor_pos_dict = {}

        if (kappa, kappa_phi) != (
            new_kappa,
            new_kappa_phi,
        ) and self.minikappa_correction_hwobj is not None:
            sampx = self.motor_hwobj_dict["sampx"].get_value()
            sampy = self.motor_hwobj_dict["sampy"].get_value()
            phiy = self.motor_hwobj_dict["phiy"].get_value()
            new_sampx, new_sampy, new_phiy = self.minikappa_correction_hwobj.shift(
                kappa, kappa_phi, [sampx, sampy, phiy], new_kappa, new_kappa_phi
            )

            motor_pos_dict[self.motor_hwobj_dict["kappa"]] = new_kappa
            motor_pos_dict[self.motor_hwobj_dict["kappa_phi"]] = new_kappa_phi
            motor_pos_dict[self.motor_hwobj_dict["sampx"]] = new_sampx
            motor_pos_dict[self.motor_hwobj_dict["sampy"]] = new_sampy
            motor_pos_dict[self.motor_hwobj_dict["phiy"]] = new_phiy

            self.move_motors(motor_pos_dict, timeout=30)

    def visual_align(self, point_1, point_2):
        """
        Descript. :
        """
        if self.in_plate_mode:
            self.log.info("PX2Diffractometer: Visual align not available in Plate mode")
        else:
            t1 = [point_1.sampx, point_1.sampy, point_1.phiy]
            t2 = [point_2.sampx, point_2.sampy, point_2.phiy]
            kappa = self.motor_hwobj_dict["kappa"].get_value()
            phi = self.motor_hwobj_dict["kappa_phi"].get_value()
            (
                new_kappa,
                new_phi,
                (
                    new_sampx,
                    new_sampy,
                    new_phiy,
                ),
            ) = self.goniometer.get_align_vector(t1, t2, kappa, phi)
            # self.minikappa_correction_hwobj.alignVector(t1,t2,kappa,phi)
            self.move_to_motors_positions(
                {
                    self.motor_hwobj_dict["kappa"]: new_kappa,
                    self.motor_hwobj_dict["kappa_phi"]: new_phi,
                    self.motor_hwobj_dict["sampx"]: new_sampx,
                    self.motor_hwobj_dict["sampy"]: new_sampy,
                    self.motor_hwobj_dict["phiy"]: new_phiy,
                }
            )

    def re_emit_values(self):
        """
        Description:
        """
        self.emit("minidiffPhaseChanged", (self.current_phase,))
        self.emit("omegaReferenceChanged", (self.reference_pos,))
        self.emit("minidiffShutterStateChanged", (self.fast_shutter_is_open,))

    def toggle_fast_shutter(self):
        """
        Description:
        """
        if self.chan_fast_shutter_is_open is not None:
            self.chan_fast_shutter_is_open.set_value(not self.fast_shutter_is_open)

    def find_loop(self):
        """
        Description:
        """
        image_array = HWR.beamline.sample_view.get_snapshot(return_as_array=True)
        (info, x, y) = lucid.find_loop(image_array)
        surface_score = 10
        return x, y, surface_score

    def move_omega_relative(self, relative_angle):
        """
        Description:
        """
        self.motor_hwobj_dict["phi"].set_value_relative(relative_angle, 5)

    def close_kappa(self):
        """
        Descript. :
        """
        gevent.spawn(self.close_kappa_task)

    def close_kappa_task(self):
        """Close kappa task"""
        self.log.debug("Started closing Kappa")
        self.move_kappa_and_phi_procedure(0, None)
        self.wait_device_ready(60)
        self.motor_hwobj_dict["kappa"].homeMotor()
        self.wait_device_ready(60)
        self.move_kappa_and_phi_procedure(0, None)
        self.wait_device_ready(60)
        self.log.debug("Done closing Kappa")
        # self.kappa_phi_motor_hwobj.homeMotor()

    def set_zoom(self, position):
        """ """
        self.zoom_motor_hwobj.moveToPosition(position)

    def get_status(self):
        self.current_status = self.chan_status.get_value()
        return self.current_status

    def get_point_from_line(self, point_one, point_two, frame_num, frame_total):
        """
        Descript. : method used to get a new motor position based on a position
                    between two positions. As arguments both motor positions are
                    given. frame_num and frame_total is used estimate new point position
                    Helical line goes from point_one to point_two.
                    In this direction also new position is estimated
        """
        new_point = {}
        point_one = point_one.as_dict()
        point_two = point_two.as_dict()
        for motor in point_one.keys():
            new_motor_pos = point_one[motor] + (
                point_two[motor] - point_one[motor]
            ) * frame_num / float(frame_total)
            new_point[motor] = new_motor_pos
        return new_point

    def get_osc_limits(self):
        return self.motor_hwobj_dict["phi"].get_dynamic_limits()

    def get_osc_max_speed(self):
        return self.motor_hwobj_dict["phi"].get_max_speed()

    def get_scan_limits(self, speed=None, num_images=None, exp_time=None):
        """optical_alignment
        Gets scan limits. Necessary for example in the plate mode
        where osc range is limited
        """
        if speed is not None:
            return self.cmd_get_omega_scan_limits(speed), None
        total_exposure_time = num_images * exp_time
        tmp = self.cmd_get_omega_scan_limits(0)
        max_speed = self.get_osc_max_speed()
        w0 = tmp[0]
        w1 = tmp[1]
        x1 = 10
        x2 = 100

        c1 = self.cmd_get_omega_scan_limits(x1)[0] - w0
        c2 = self.cmd_get_omoptical_alignmentega_scan_limits(x2)[0] - w0

        a = -(c2 * x1 - c1 * x2) / (x1 * x2 * (x1 - x2))
        b = -(-c2 * pow(x1, 2) + c1 * pow(x2, 2)) / (x1 * x2 * (x1 - x2))

        result_speed = (
            (
                -2 * b
                - total_exposure_time
                + sqrt((2 * b + total_exposure_time) ** 2 - 8 * a * (w0 - w1))
            )
            / 4
            / a
        )
        if result_speed < 0:
            return (None, None), None
        elif result_speed > max_speed:
            delta = a * max_speed**2 + b * max_speed
            # total_exposure_time = total_exposure_time * result_speed / max_speed
            total_exposure_time = (w1 - w0 - 2 * delta) / (max_speed - 0.1)
        else:
            delta = a * result_speed**2 + b * result_speed

        return (w0 + delta, w1 - delta), total_exposure_time / num_images

        """
        if speed is not None:
            return self.cmd_get_omega_scan_limits(speed)
        elif speed == 0:
            return self.get_osc_limits()
        else:
            motor_acc_const = 5
            motor_acc_time = num_images / exp_time / motor_acc_const
            min_acc_time = 0.0015
            acc_time = max(motor_acc_time, min_acc_time)

            shutter_time = 3.7 / 1000.
            max_limit = num_images / exp_time * (acc_time+2*shutter_time + 0.2) / 2

            return [0, max_limit]
        """

    def get_scintillator_position(self):
        return self.chan_scintillator_position.get_value()

    def set_scintillator_position(self, position):
        self.chan_scintillator_position.set_value(position)
        with gevent.Timeout(5, Exception("Timeout waiting for scintillator position")):
            while position != self.get_scintillator_position():
                gevent.sleep(0.01)

    def get_capillary_position(self):
        return self.chan_capillary_position.get_value()

    def set_capillary_position(self, position):
        self.chan_capillary_position.set_value(position)
        with gevent.Timeout(5, Exception("Timeout waiting for capillary position")):
            while position != self.get_capillary_position():
                gevent.sleep(0.01)

    def zoom_in(self):
        self.zoom_motor_hwobj.zoom_in()

    def zoom_out(self):
        self.zoom_motor_hwobj.zoom_out()

    def save_centring_positions(self):
        self.cmd_save_centring_positions()

    def beam_position_check(self):
        logging.getLogger("user_level_log").info("Going to check the beam position")
        self.bpc(wait=False)

    @task
    def bpc(self):
        ba = beam_align.beam_align(
            name_pattern="%s_%s" % (os.getuid(), time.asctime().replace(" ", "_")),
            directory="%s/beam_align" % os.getenv("HOME"),
        )
        logging.getLogger("user_level_log").info(
            "Align beam to the optical centre of the camera"
        )
        logging.getLogger("user_level_log").info(
            "Moving scintillator to sample position, please wait ..."
        )
        ba.execute()

        if ba.no_beam is False:
            logging.getLogger("user_level_log").info(
                "Initial mirror positions (vfm, hfm) [mrad]: %.4f %.4f"
                % tuple(ba.initial_mirror_positions)
            )
            logging.getLogger("user_level_log").info(
                "Initial pixel shift from center (vertical, horizontal): %.1f, %.1f"
                % tuple(ba.initial_pixel_shift)
            )
            logging.getLogger("user_level_log").info(
                "Beam position adjustment finished after %d iterations"
                % ba.number_of_iterations
            )
            logging.getLogger("user_level_log").info(
                "Final mirror positions (vfm, hfm) [mrad]: %.4f %.4f"
                % tuple(ba.final_mirror_position)
            )
            logging.getLogger("user_level_log").info(
                "Final pixel shift from center (vertical, horizontal): %.1f, %.1f"
                % tuple(ba.final_pixel_shift)
            )
            logging.getLogger("user_level_log").info(
                "Delta in motor positions [mrad]: %.4f, %.4f"
                % tuple(ba.final_mirror_position - ba.initial_mirror_positions)
            )
        else:
            logging.getLogger("user_level_log").info(ba.no_beam_message)

    @task
    def anneal(self, time=1.0):
        anneal_procedure(time)

    @task
    def excenter(
        self,
        scan_length=0.1,
        step=90.0,
        start=0.0,
        base_directory="/nfs/ruche/proxima2a-spool/2019_Run1/excenter",
        name_pattern="excenter",
    ):
        directory = os.path.join(base_directory, datetime.datetime.today().isoformat())

        angles = str(tuple(np.arange(start, 360.0, step)))

        execute_line = 'excenter.py -d %s -n %s -l %.2f -a "%s" &' % (
            directory,
            name_pattern,
            scan_length,
            angles,
        )

        self.log.info("excenter angles %s" % angles)
        self.log.info("excenter line %s" % execute_line)

        os.system(execute_line)

    def aperture_align(self):
        logging.getLogger("user_level_log").info("Aligning the current aperture")
        self.aa(wait=False)

    @task
    def aa(self):
        logging.getLogger("user_level_log").info(
            "Adjusting camera exposure time for visualisation on the scintillator"
        )
        a = scan_and_align.scan_and_align("aperture", display=False)
        logging.getLogger("user_level_log").info("Scanning the aperture")
        a.scan()
        a.align(optimum="com")
        a.save_scan()
        logging.getLogger("user_level_log").info(
            "Setting camera exposure time back to 0.050 seconds"
        )
        logging.getLogger("user_level_log").info("Aligning aperture finished")
        a.predict()

    def start_automatic_centring(
        self, sample_info=None, loop_only=False, wait_result=None
    ):
        """ """
        self.emit_progress_message("Automatic centring...")
        self.current_centring_procedure = gevent.spawn(self.optical_alignment)
        self.current_centring_procedure.link(self.centring_done)

        if wait_result:
            self.ready_event.wait()
            self.ready_event.clear()

        # self.optical_alignment(wait=False)

    # @task
    def optical_alignment(self):
        start = time.time()
        oa = optical_alignment.optical_alignment(
            name_pattern="%s_%s" % (os.getuid(), time.asctime().replace(" ", "_")),
            directory="%s/automated_optical_alignment" % os.getenv("HOME"),
            analysis=True,
            conclusion=True,
            move_zoom=True,
            film_step=120.0,
        )

        oa.execute()

        self.log.info(
            "optical sample alignment finished in %.3f seconds" % (time.time() - start)
        )
        # self.emit_centring_successful()
        self.emit("centringSuccessful", self.current_centring_method, ())
        result_position = oa.get_result_position()
        translated_position = self.translate_from_md2_to_mxcube(result_position)
        return translated_position

    def set_nclicks(self, nclicks):
        self.log.info(
            "PX2Diffractometer: number of centring clicks changed: %s" % nclicks
        )
        try:
            self.nclicks = int(nclicks)
        except Exception:
            self.log.exception("")

    def set_step(self, step):
        self.log.info("PX2Diffractometer: centring step changed: %s" % step)
        try:
            self.step = float(step)
        except Exception:
            self.log.exception("")

    def set_centring_method(self, centring_method):
        self.log.info(
            "PX2Diffractometer: centring method changed: %s" % centring_method
        )
        try:
            self.centring_method = centring_method
        except Exception:
            self.log.exception("")

    def is_ready(self):
        """
        Detects if device is ready
        """
        condition1 = self.current_state == DiffractometerState.tostring(
            DiffractometerState.Ready
        )

        condition2 = self.collecting == False

        # self.log.info('Diffractometer is_ready collecting %s' % self.collecting)

        return condition1 and condition2

    def set_collecting(self, collecting=True):
        self.collecting = collecting



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Collect procedures, using bliss for scanning
Example xml configuration:

.. code-block:: xml

 <object class="ESRF.ESRFMultiCollect">
   <username>Photon flux</username>
   <object role="controller" href="/bliss"/>
   <scan_object>md2scan</scan_object>
   <num_snapshots>4</num_snapshots>
   <geometry>[-1, 0, 0, 0, -1, 0]</geometry>
 </object>
"""

import logging
import os
import shutil
from ast import literal_eval
from pathlib import Path

from lima2mxh5master.api import convert_lima2_to_h5mx

from mxcubecore import HardwareRepository as HWR

from .ESRFMultiCollect import ESRFMultiCollect
from .fill_meta_data import FillMetaData


class BlissMultiCollect(ESRFMultiCollect):
    def __init__(self, name):
        super().__init__(name)
        self.fast_characterisation = None
        self.geometry = []
        self.metadata = None
        self._data_collect_parameters = None
        self.last_bliss_scan = None

    def init(self):
        super().init()
        # self.number_of_snapshots = self.get_property("num_snapshots", 4)
        _bliss = self.get_object_by_role("controller")
        _name = self.get_property("scan_object")

        # this is the geometry of the diffractometer
        try:
            self.geometry = literal_eval(self.get_property("geometry"))
        except ValueError:
            self.geometry = [-1, 0, 0, 0, -1, 0]
        self._scan = getattr(_bliss, _name)
        self.metadata = FillMetaData()

    def data_collection_hook(self, data_collect_parameters):
        pass

    def data_collection_end_hook(self, data_collect_parameters):
        self._detector._emit_status()  # noqa: SLF001

        oscillation_parameters = data_collect_parameters["oscillation_sequence"][0]
        subwedge_size = oscillation_parameters.get("reference_interval", 1)
        wedges_to_collect = self.prepare_wedges_to_collect(
            oscillation_parameters["start"],
            oscillation_parameters["number_of_images"],
            oscillation_parameters["range"],
            subwedge_size,
            oscillation_parameters["overlap"],
        )

        run_number = data_collect_parameters["fileinfo"]["run_number"]

        for wedge_number in range(1, len(wedges_to_collect) + 1):
            convert_lima2_to_h5mx(
                Path(self.last_bliss_scan.scan_saving.data_fullpath),
                Path(
                    self.last_bliss_scan.scan_saving.get_path(),
                    data_collect_parameters["fileinfo"]["prefix"]
                    + "_%s_%s_master.h5" % (run_number, wedge_number),
                ),
                "%s.1" % wedge_number,
            )

    def _bliss_data_collection_hook(self, data_collect_parameters):
        # first set the proposal name for bliss
        _as = HWR.beamline.lims.get_active_session()
        proposal = f"{_as.code}{_as.number}"
        self._scan.set_session(proposal)

        sample_name = data_collect_parameters["sample_reference"]["sample_name"]
        meta_data = {
            "sample_name": sample_name.replace(":", "-"),
            "acronym": data_collect_parameters["sample_reference"]["acronym"],
            "subdir": data_collect_parameters["fileinfo"]["directory"].split(
                "RAW_DATA/"
            )[1],
        }

        # fill in the static metadata
        self._scan.prepare_data_saving(meta_data)
        self.metadata.diffractometer_static_data()
        self.metadata.detector_static_data()

        try:
            comment = HWR.beamline.sample_changer.get_crystal_id()
            data_collect_parameters["comment"] = comment
        except AttributeError:
            self.log.exception()

    def last_image_saved(self, total_time, exptime, num_images):
        if self._scan._lima_object.tango_ctrl_dev.acq_state == "running":  # noqa: SLF001
            # if HWR.beamline.detector.status["acq_satus"] == "RUNNING":
            return int(total_time / exptime)
        # return HWR.beamline.detector.last_image_saved()
        return num_images

    def get_beam_size(self):
        _width, _height, _, _ = HWR.beamline.beam.get_value()
        return _width, _height

    def get_slit_gaps(self):
        return (None, None)

    def get_beam_shape(self):
        return HWR.beamline.beam.get_value()[2].name

    def get_resolution_at_corner(self):
        return HWR.beamline.resolution.get_value_at_corner()

    def ready(*motors):
        return not any(m.motorIsMoving() for m in motors)

    def move_motors(self, motors_to_move_dict):
        # We do not want to modify the input dict
        motor_positions_copy = motors_to_move_dict.copy()
        diffr = HWR.beamline.diffractometer
        for tag in ("kappa", "kappa_phi", "zoom"):
            if tag in motor_positions_copy:
                del motor_positions_copy[tag]

        diffr.move_sync_motors(motor_positions_copy, wait=True, timeout=200)

    def take_crystal_snapshots(self, number_of_snapshots, image_path_list=None):
        if image_path_list is None:
            image_path_list = []
        HWR.beamline.diffractometer.take_snapshot(image_path_list)

    def do_prepare_oscillation(self, *args, **kwargs):
        diffr = HWR.beamline.diffractometer
        # set the detector cover out
        try:
            diffr.open_detector_cover()
        except Exception:
            logging.getLogger("HWR").exception("Could not open detector cover")

        # send again the command as MD2 software only handles one
        # centered position!!
        # has to be where the motors are and before changing the phase
        # diffr.get_command_object("save_centring_positions")()

        # switch on the front light
        front_light_switch = diffr.get_object_by_role("FrontLightSwitch")
        front_light_switch.set_value(front_light_switch.VALUES.IN)
        # diffr.get_object_by_role("FrontLight").set_value(2)

        # move to DataCollection phase
        logging.getLogger("user_level_log").info("Moving MD2 to DataCollection")
        # AB next line to speed up the data collection
        diffr.set_phase("DataCollection", wait=False, timeout=0)

    def data_collection_cleanup(self):
        self.close_fast_shutter()

    def diffractometer_metadata(self, start, end, exptime, nb_images):
        """Prepare the diffractometer related meta data"."""
        self.metadata.diffractometer_static_data()
        self.metadata.geometry["orientation"]["value"] = self.geometry

    def detector_metadata(self, start, end, exptime, nb_images):
        """Prepare the detector related meta data"."""
        self.metadata.detector_static_data()
        self.metadata.detector_specific["x_pixels_in_detector"] = (
            HWR.beamline.detector.get_width()
        )
        self.metadata.detector_specific["y_pixels_in_detector"] = (
            HWR.beamline.detector.get_height()
        )
        self.metadata.detector_specific["photon_energy"] = (
            HWR.beamline.energy.get_value()
        )

        self.metadata.detector["detector_specific"] = self.metadata.detector_specific

        pixel_size_x, pixel_size_y = HWR.beamline.detector.get_pixel_size()
        beam_x, beam_y = HWR.beamline.detector.get_beam_position()
        readout_time = HWR.beamline.detector.get_property("deadtime")

        self.metadata.detector.update(
            {
                "detector": {
                    "beam_center_x": beam_x,
                    "beam_center_x@units": "pixel",
                    "beam_center_y": beam_y,
                    "beam_center_y@units": "pixel",
                    "count_time": exptime / nb_images,
                    "count_time@units": "s",
                    "frame_time": (exptime / nb_images) + readout_time,
                    "readout_time": readout_time,
                    "readout_time@units": "s",
                    "frame_time@units": "s",
                    "distance": HWR.beamline.detector.distance.get_value() / 1000.0,
                    "distance@units": "m",
                    "nimages": nb_images,
                    "x_pixel_size": pixel_size_x / 1000.0,
                    "x_pixel_size@units": "m",
                    "y_pixel_size": pixel_size_y / 1000.0,
                    "y_pixel_size@units": "m",
                    "counter_name": "pilatus4_4m_lima2",
                    "detector_number": "D029099",
                }
            }
        )

    def sample_metadata(self, start, end, exptime, nb_images) -> dict:
        """Prepare the meta data for the sample"""
        omega = []
        omega_end = []
        osc_range_total = abs(end - start)
        osc_range = osc_range_total / nb_images
        for img in range(nb_images):
            omega.append(start + img * osc_range)
            omega_end.append(start + (img + 1) * osc_range)

        self.metadata.sample_transformations.update(
            {
                "transformations": {
                    "omega": omega,
                    "omega_end": omega_end,
                    "omega_range_average": osc_range,
                    "omega_range_total": osc_range_total,
                    "omega@vector": self.geometry[:3],
                    "omega@unit": "deg",
                    "omega_end@unit": "deg",
                    "omega_range_average@unit": "deg",
                    "omega_range_total@unit": "deg",
                    "translation": [0],
                }
            }
        )
        self.metadata.sample_goniometer.update(
            {
                "goniometer": {
                    "omega": omega,
                    "omega_end": omega_end,
                    "omega_range_average": osc_range,
                    "omega_range_total": osc_range_total,
                }
            }
        )

    def beam_metadata(self):
        """Meta data for the beam"""
        self.metadata.beam.update(
            {
                "beam": {
                    "incident_wavelength": HWR.beamline.energy.get_wavelength(),
                    "incident_wavelength@units": "angstrom",
                }
            }
        )

    def oscil(self, start, end, exptime, nb_images, wait=True):
        """run bliss scan according to the type"""

        meta_data = {}
        self.sample_metadata(start, end, exptime, nb_images)
        self.beam_metadata()
        self.detector_metadata(start, end, exptime, nb_images)
        # self.diffractometer_metadata(start, end, exptime, nb_images)
        meta_data = self.metadata

        if self.helical:
            # helical scan
            motor_pos = self.helical_pos
            self.last_bliss_scan = self._scan.line_scan(
                start, end, exptime, nb_images, motor_pos, meta_data
            )
        elif self.mesh:
            # ??? nb_images = self.mesh_total_nb_frames
            # move the motorts to the centre of the mesh first
            HWR.beamline.diffractometer.move_motors(self.mesh_center.as_dict())
            self.last_bliss_scan = self._scan.mesh_scan(
                start,
                end,
                exptime,
                nb_images,
                self.mesh_num_lines,
                self.mesh_range,
                meta_data,
            )
        elif self.fast_characterisation:
            self.nb_frames = 10
            self.nb_scan = 4
            self.angle = 90
            exptime *= 10
            _end = (end - start) * 10
            self.last_bliss_scan = self._scan.characterisation_scan(
                start,
                _end,
                exptime,
                self.nb_frames,
                self.nb_scan,
                self.angle,
                meta_data,
            )
        else:
            self.last_bliss_scan = self._scan.osc_scan(
                start, end, exptime, nb_images, meta_data
            )

    def prepare_acquisition(
        self, take_dark, start, osc_range, exptime, npass, number_of_images, comment=""
    ):
        pass

    def start_acquisition(self, exptime, npass, first_frame, shutterless):
        pass

    def open_fast_shutter(self):
        fs = HWR.beamline.fast_shutter
        fs.set_value(fs.VALUES.OPEN)

    def close_fast_shutter(self):
        fs = HWR.beamline.fast_shutter
        fs.set_value(fs.VALUES.CLOSED)

    def set_helical(self, helical_on):
        self.helical = helical_on

    def set_helical_pos(self, helical_oscil_pos):
        self.helical_pos = helical_oscil_pos

    # specifies the next scan will be a mesh scan
    def set_mesh(self, mesh_on):
        self.mesh = mesh_on

    def set_mesh_scan_parameters(
        self, num_lines, total_nb_frames, mesh_center_param, mesh_range_param
    ):
        """
        sets the mesh scan parameters :
         - vertcal range
         - horizontal range
         - nb lines
         - nb frames per line
        """
        self.mesh_num_lines = num_lines
        self.mesh_total_nb_frames = total_nb_frames
        self.mesh_range = mesh_range_param
        self.mesh_center = mesh_center_param

    def set_fast_characterisation(self, value=False):
        self.fast_characterisation = value

    def get_cryo_temperature(self):
        return 0

    def prepare_intensity_monitors(self):
        return

    def get_beam_centre(self):
        pixel_x, pixel_y = HWR.beamline.detector.get_pixel_size()
        bcx, bcy = HWR.beamline.detector.get_beam_position()
        return [bcx * pixel_x, bcy * pixel_y]

    def write_input_files(self, datacollection_id):
        # copy *geo_corr.cbf* files to process directory
        try:
            process_dir = os.path.join(self.xds_directory, "..")
            raw_process_dir = os.path.join(self.raw_data_input_file_dir, "..")
            for _dir in (process_dir, raw_process_dir):
                for filename in ("x_geo_corr.cbf.bz2", "y_geo_corr.cbf.bz2"):
                    dest = os.path.join(_dir, filename)
                    if os.path.exists(dest):
                        continue
                    shutil.copyfile(
                        os.path.join(
                            self.get_property("template_file_directory"), filename
                        ),
                        dest,
                    )
        except Exception:
            logging.exception("Exception happened while copying geo_corr files")

        return super().write_input_files(datacollection_id)



import time

import gevent
import numpy

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.TaskUtils import cleanup

SCAN_LENGTH = 500


class XRFMockup(HardwareObject):
    def __init__(self, name):
        super().__init__(name)

    def init(self):
        self.ready_event = gevent.event.Event()
        self.spectrumInfo = {}
        self.spectrumInfo["startTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
        self.spectrumInfo["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
        self.spectrumInfo["energy"] = 0
        self.spectrumInfo["beamSizeHorizontal"] = 0
        self.spectrumInfo["beamSizeVertical"] = 0
        self.ready_event = gevent.event.Event()
        self.__scan_data = {}

        # self.plottin_hwobj = self.get_object_by_role('plotting')

    def is_connected(self):
        return True

    def canSpectrum(self):
        return True

    def gaussian(self, x, mu, sig):
        return numpy.exp(-numpy.power(x - mu, 2.0) / (2 * numpy.power(sig, 2.0)))

    # for mu, sig in [(-1, 1), (0, 2), (2, 3)]:
    #     mp.plot(gaussian(np.linspace(-3, 3, 120), mu, sig))

    def startXrfSpectrum(
        self, ct, directory, arch_dir, prefix, session_id=None, blsample_id=None
    ):
        s1 = self.gaussian(numpy.linspace(0, 3, SCAN_LENGTH), 1, 0.1)
        s2 = self.gaussian(numpy.linspace(0, 3, SCAN_LENGTH), 2, 0.3) * (
            0.25 * numpy.random.rand()
        )
        raw_data = s1 + s2

        self.log.info("XRF Spectrum Started, task id: %d" % blsample_id)
        self.scanning = True
        self.emit("xrfSpectrumStarted", ())
        with cleanup(self.ready_event.set):
            self.spectrumInfo["sessionId"] = session_id
            self.spectrumInfo["blSampleId"] = blsample_id
            if blsample_id is None:
                blsample_id = numpy.random.randint(1, 999999)

            scan_info = {
                "scan_nb": blsample_id,
                "title": "XRF Scan",
                "labels": ["energy", "diode value"],
            }
            scan_id = scan_info["scan_nb"]
            self.__scan_data[scan_id] = []

            self.emit(
                "new_plot",
                {
                    "id": scan_info["scan_nb"],
                    "title": scan_info["title"],
                    "labels": scan_info["labels"],
                },
            )

            for i in range(SCAN_LENGTH):
                try:
                    data = {"energy": i, "diode value": raw_data[i]}
                    new_data = numpy.column_stack(
                        [data[name] for name in scan_info["labels"]]
                    )
                    self.__scan_data[scan_id].append(new_data)
                    aux = numpy.concatenate(self.__scan_data[scan_id]).tolist()
                    self.emit("plot_data", {"id": scan_id, "data": aux})
                    if divmod(i, SCAN_LENGTH / 10)[1] == 0:
                        progress = i / float(SCAN_LENGTH)
                        self.log.info("XRF Spectrum Progress %f" % progress)
                        self.emit("xrf_task_progress", (blsample_id, progress))

                    gevent.sleep(0.02)
                except Exception as ex:
                    print(("Exception ", ex))

            self.emit(
                "plot_end",
                {
                    "id": scan_id,
                    "data": numpy.concatenate(self.__scan_data[scan_id]).tolist(),
                    "type": "XRFScan",
                },
            )

            mcaCalib = [10, 1, 21, 0]
            mcaConfig = {}
            mcaConfig["legend"] = "XRF test scan from XRF mockup"
            mcaConfig["htmldir"] = "html dir not defined"
            mcaConfig["min"] = raw_data[0]
            mcaConfig["max"] = raw_data[-1]
            mcaConfig["file"] = None
            res = []
            for arr in self.__scan_data[scan_id]:
                res.append(arr[0].tolist())

            self.emit("xrfSpectrumFinished", (res, mcaCalib, mcaConfig))
            self.log.info("XRF Spectrum Finished")
            del self.__scan_data[scan_id]



import json
import logging
import math

import gevent
from devtools import debug
from pydantic.v1 import Field

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.ESRF.queue_entry.ssx_base_queue_entry import (
    BaseUserCollectionParameters,
    SsxBaseQueueEntry,
    SsxBaseQueueTaskParameters,
    SSXPathParameters,
)
from mxcubecore.model.common import (
    CommonCollectionParamters,
    LegacyParameters,
    StandardCollectionParameters,
)
from mxcubecore.model.queue_model_objects import DataCollection

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class SSXUserCollectionParameters(BaseUserCollectionParameters):
    num_images: int = Field(0, description="")
    horizontal_spacing: float = Field(20, gt=0, lt=1000, unit="um")
    vertical_spacing: float = Field(20, gt=0, lt=1000, unit="um")

    #    _chip_name_tuple = tuple(
    #        HWR.beamline.diffractometer.get_head_configuration().available.keys()
    #    )
    #    _current_chip = HWR.beamline.diffractometer.get_head_configuration().current
    #    chip_type: Literal[_chip_name_tuple] = Field(_current_chip)

    class Config:
        extra = "ignore"


class SsxCustomFoilCollectionTaskParameters(SsxBaseQueueTaskParameters):
    path_parameters: SSXPathParameters
    common_parameters: CommonCollectionParamters
    collection_parameters: StandardCollectionParameters
    user_collection_parameters: SSXUserCollectionParameters
    legacy_parameters: LegacyParameters

    @staticmethod
    def ui_schema():
        schema = json.loads(SsxBaseQueueTaskParameters.ui_schema())
        schema.update(
            {
                "sub_sampling": {"ui:readonly": "true"},
            }
        )
        return json.dumps(schema)

    @staticmethod
    def update_dependent_fields(field_data):
        horizontal_spacing = field_data.get("horizontal_spacing", 0)
        vertical_spacing = field_data.get("vertical_spacing", 0)
        sub_sampling = field_data["sub_sampling"]

        num_images, _, _ = (
            SsxCustomFoilCollectionTaskParameters.calculate_number_of_images(
                horizontal_spacing, vertical_spacing, sub_sampling
            )
        )

        return {"num_images": num_images}

    @staticmethod
    def calculate_number_of_images(horizontal_spacing, vertical_spacing, sub_sampling):
        chip_data = HWR.beamline.diffractometer.get_head_configuration().available[
            "CUSTOM_FOIL"
        ]

        chip_width = (
            chip_data.sections[0].calibration_data.top_right[0]
            - chip_data.sections[0].calibration_data.top_left[0]
        )
        chip_height = (
            chip_data.sections[0].calibration_data.bottom_left[1]
            - chip_data.sections[0].calibration_data.top_left[1]
        )

        nb_samples_per_line = math.floor(
            chip_width / ((horizontal_spacing / 1000) * sub_sampling)
        )
        nb_lines = math.floor(chip_height / (vertical_spacing / 1000))

        num_images = (
            math.floor((nb_samples_per_line * nb_lines) / 2)
            * 2
            * len(chip_data.sections)
        )

        return num_images, nb_lines, nb_samples_per_line


class SsxCustomFoilCollectionQueueModel(DataCollection):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class SsxCustomFoilCollectionQueueEntry(SsxBaseQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    QMO = SsxCustomFoilCollectionQueueModel
    DATA_MODEL = SsxCustomFoilCollectionTaskParameters
    NAME = "SSX Foil LLNL Collection"
    REQUIRES = ["point", "line", "no_shape", "chip", "grid"]

    # New style queue entry does not take view argument,
    # adding kwargs for compatability, but they are unsued
    def __init__(self, view, data_model: SsxCustomFoilCollectionQueueModel):
        super().__init__(view=view, data_model=data_model)
        params = self._data_model._task_data.user_collection_parameters
        (
            num_images,
            _nb_lines,
            _nb_samples_per_line,
        ) = SsxCustomFoilCollectionTaskParameters.calculate_number_of_images(
            params.horizontal_spacing, params.vertical_spacing, params.sub_sampling
        )
        self._data_model._task_data.collection_parameters.num_images = num_images
        self._data_model._task_data.user_collection_parameters.num_images = num_images

    def execute(self):
        super().execute()

        debug(self._data_model._task_data)

        enforce_centring_phase = False

        params = self._data_model._task_data.user_collection_parameters
        exp_time = params.exp_time
        chip_data = HWR.beamline.diffractometer.get_head_configuration().available[
            "CUSTOM_FOIL"
        ]

        fname_prefix = self._data_model._task_data.path_parameters.prefix
        data_root_path = self.get_data_path()
        reject_empty_frames = (
            self._data_model._task_data.user_collection_parameters.reject_empty_frames
        )

        (
            num_images,
            nb_lines,
            nb_samples_per_line,
        ) = SsxCustomFoilCollectionTaskParameters.calculate_number_of_images(
            params.horizontal_spacing, params.vertical_spacing, params.sub_sampling
        )

        HWR.beamline.diffractometer.wait_ready()
        HWR.beamline.diffractometer.set_phase("DataCollection")

        self.take_pedestal()

        logging.getLogger("user_level_log").info("Preparing detector")
        HWR.beamline.detector.prepare_acquisition(
            num_images,
            exp_time,
            data_root_path,
            fname_prefix,
            dense_skip_nohits=reject_empty_frames,
        )

        fname_prefix = self._data_model._task_data.path_parameters.prefix
        fname_prefix += "_foil_"

        if HWR.beamline.control.safshut_oh2.state.name != "OPEN":
            logging.getLogger("user_level_log").info("Opening OH2 safety shutter")
            HWR.beamline.control.safshut_oh2.open()

        HWR.beamline.detector.wait_ready()
        HWR.beamline.detector.start_acquisition()
        logging.getLogger("user_level_log").info(
            "Detector ready, waiting for trigger ..."
        )

        for section in chip_data.sections:
            logging.getLogger("user_level_log").info(
                f"Collecting region {chip_data.sections.index(section) + 1} of {len(chip_data.sections)}"
            )
            region = (
                section.calibration_data.top_left[0],
                section.calibration_data.top_left[1],
                section.calibration_data.top_left[2],
                section.calibration_data.top_right[0],
                section.calibration_data.top_right[1],
                section.calibration_data.top_right[2],
                section.calibration_data.bottom_left[0],
                section.calibration_data.bottom_left[1],
                section.calibration_data.bottom_left[2],
            )

            self.start_processing("FOIL")

            logging.getLogger("user_level_log").info(f"Defining region {region}")

            HWR.beamline.diffractometer.prepare_ssx_grid_scan(
                *region, nb_samples_per_line, nb_lines
            )

            HWR.beamline.diffractometer.wait_ready()

            logging.getLogger("user_level_log").info(f"Acquiring region {region}")
            logging.getLogger("user_level_log").info(
                f"Sub sampling is {params.sub_sampling}"
            )
            logging.getLogger("user_level_log").info(
                f"Acquiring {num_images / len(chip_data.sections)} images ({nb_lines} lines x {nb_samples_per_line} samples per line)"
            )
            logging.getLogger("user_level_log").info(
                f"Data path: {data_root_path}{fname_prefix}*.h5"
            )

            try:
                HWR.beamline.diffractometer.start_ssx_scan(enforce_centring_phase)
            except Exception:
                msg = "Diffractometer start failed! Stopping the detector"
                logging.getLogger("user_level_log").exception(msg)
                HWR.beamline.detector.stop_acquisition()
                return

            logging.getLogger("user_level_log").info("Waiting for scan to finish ...")
            HWR.beamline.diffractometer.wait_ready()
            logging.getLogger("user_level_log").info("Scan finished ...")
            logging.getLogger("user_level_log").info(f"Acquired {region}")

        try:
            HWR.beamline.diffractometer.wait_ready()
        finally:
            HWR.beamline.detector.wait_ready()
            acquired = HWR.beamline.detector.get_acquired_frames()
            logging.getLogger("user_level_log").info(
                f"Acquired {len(chip_data.sections)} regions"
            )
            logging.getLogger("user_level_log").info(
                f"Acquired total: {acquired} images"
            )

            HWR.beamline.diffractometer.wait_ready()
            HWR.beamline.diffractometer.set_phase("Transfer", wait=True, timeout=120)
            logging.getLogger("user_level_log").info("set to Transfer phase")

    def pre_execute(self):
        super().pre_execute()

    def post_execute(self):
        super().post_execute()

    def stop(self):
        logging.getLogger("user_level_log").info("Stopping diffractometer ...")
        HWR.beamline.diffractometer.abort_cmd()
        gevent.sleep(5)
        HWR.beamline.diffractometer.wait_ready()

        super().stop()



import time
from typing import Optional

import gevent
import numpy as np

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractActuator import AbstractActuator


class EPICSActuator(AbstractActuator):
    ACTUATOR_VAL = "val"  # setpoint
    ACTUATOR_RBV = "rbv"  # readback

    def __init__(self, name):
        super().__init__(name)
        self.setpoint = None
        self._nominal_limits = (-1e4, 1e4)
        self.default_timeout = 180
        if not self.unit:
            self.unit = 10**-3

    def init(self):
        super().init()
        self.update_state(self.STATES.READY)
        self.connect(self.get_channel_object("rbv"), "update", self.update_value)

    def hasnt_arrived(self, setpoint):
        readback = self.get_value()
        if not readback:
            return False
        return not np.isclose(readback, setpoint, rtol=self.unit, atol=self.unit)

    def wait_ready(self, timeout: Optional[float] = None):
        self._ready_event.clear()
        is_set = self._ready_event.is_set()
        try:
            with gevent.Timeout(timeout, exception=TimeoutError):
                while not is_set:
                    is_set = self._ready_event.is_set()
                    if not self.hasnt_arrived(self.setpoint):
                        self._ready_event.set()
                    time.sleep(0.15)
        except TimeoutError:
            pvname = self.get_channel_object("").command.pv_name
            self.print_log(
                level="error",
                msg=f"{pvname} motion has timed out.",
            )
        self.update_state(self.STATES.READY)

    def get_value(self):
        return self.get_channel_value(self.ACTUATOR_RBV)

    def _set_value(self, value):
        self.setpoint = value
        self.update_state(self.STATES.BUSY)
        self.set_channel_value(self.ACTUATOR_VAL, value)

    def set_value(self, value, timeout: float = 0):
        if not timeout:
            timeout = self.default_timeout
        try:
            super().set_value(value, timeout)
        except ValueError:
            pvname = self.get_channel_object("rbv").command.pv_name
            msg = f"{pvname} value {value} is outside of actuator limits"
            msg += f": {self._nominal_limits}"
            self.print_log(
                level="error",
                msg=msg,
            )

    def abort(self):
        if self._wait_task is not None:
            self._wait_task.set()
        self.update_state(self.STATES.READY)


class EPICSActuatorBluesky(EPICSActuator):
    """
    This class alters the _set_value function of EPICSActuator for
    cases when the set is done via bluesky rather than directly by
    MXCuBE. The plan's name and parameter must be specified at the
    configuration file. Because wait_ready function works the same
    way as EPICSActuator, frontend functionality remains the same.

    YAML Example
    ------------

    %YAML 1.2
    ---
    class: LNLS.EPICS.EPICSActuator.EPICSActuatorBluesky
    epics:
    "MNC:A:DCM01:":
        channels:
        rbv:
            suffix: "GonRx_Energy_RBV"
            polling_period: 200
        val:
            suffix: "Energy_SP"
    configuration:
    tolerance: 0.01
    plan_name: "move_energy_and_phase"
    plan_parameter: "energy"
    default_limits: (5, 20)
    """

    def __init__(self, name):
        super().__init__(name)
        self._bluesky_api = HWR.beamline.get_object_by_role("bluesky")

    def init(self):
        super().init()
        self.plan_name = self.get_property("plan_name")
        self.plan_parameter = self.get_property("plan_parameter")

    def _set_value(self, value):
        self.setpoint = value
        self.update_state(self.STATES.BUSY)
        self._bluesky_api.execute_plan(
            plan_name=self.plan_name,
            kwargs={
                self.plan_parameter: value,
            },
        )



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
[Name]
QtLimaVideo

[Description]
HwObj used to grab images via LImA library or Lima Tango server.

[Configuration]
See example below.  To select between "Library" or "Tango" simply
use and configure the field <address> (for Library)
or <tangoname> (for Tango)
in the XML file.


Example Hardware Object XML file :
==================================
<object class="QtLimaVideo">
   <type>basler</type>
   <encoding>yuv422p</encoding>
   <address>84.89.227.6</address>
   <gain>0.5</gain>
   <exposure>0.01</exposure>
   <mirror>(False, False)</mirror>
   <interval>30</interval>
</object>
"""

from mxcubecore.HardwareObjects.abstract.AbstractVideoDevice import AbstractVideoDevice

try:
    from Lima import Core
except Exception:
    pass

try:
    from Lima import Prosilica
except ImportError as e:
    pass

try:
    from Lima import Basler
except ImportError as e:
    pass


class QtLimaVideo(AbstractVideoDevice):
    """
    Descript. :
    """

    def __init__(self, name):
        """
        Descript. :
        """
        AbstractVideoDevice.__init__(self, name)

        self.cam_address = None

        # LIMA access
        self.camera = None
        self.interface = None
        self.control = None
        self.video = None
        self.master_mode = True

    def init(self):
        self.cam_address = self.get_property("address")
        self.cam_type = self.get_property("type").lower()

        if self.cam_type == "prosilica":
            self.camera = Prosilica.Camera(self.cam_address, self.master_mode, False)
            self.interface = Prosilica.Interface(self.camera)
        elif self.cam_type == "basler":
            self.log.info("Connecting to camera with address %s" % self.cam_address)
            self.camera = Basler.Camera(self.cam_address)
            self.interface = Basler.Interface(self.camera)

        self.control = Core.CtControl(self.interface)
        self.video = self.control.video()

        AbstractVideoDevice.init(self)

    def set_cam_encoding(self, cam_encoding):
        if cam_encoding == "yuv422p":
            self.video.setMode(Core.YUV422)
        elif cam_encoding == "y8":
            self.video.setMode(Core.Y8)

        AbstractVideoDevice.set_cam_encoding(self, cam_encoding)

    """ Overloading of AbstractVideoDevice methods """

    def get_raw_image_size(self):
        if self.cam_type == "prosilica":
            return list(self.camera.getMaxWidthHeight())
        elif self.cam_type == "basler":
            width = self.camera.getRoi().getSize().get_width()
            height = self.camera.getRoi().getSize().get_height()
            return [width, height]
        else:
            return [None, None]

    def get_image(self):
        image = self.video.getLastImage()
        if image.frameNumber() > -1:
            raw_buffer = image.buffer()
            return raw_buffer, image.width(), image.height()
        else:
            return None, None, None

    def get_gain(self):
        value = self.video.getGain()
        return value

    def set_gain(self, gain_value):
        if self.master_mode:
            self.video.setGain(gain_value)

    def get_exposure_time(self):
        return self.video.getExposure()

    def set_exposure_time(self, exposure_time_value):
        if self.master_mode:
            self.video.setExposure(exposure_time_value)

    def get_video_live(self):
        return self.video.getLive()

    def set_video_live(self, flag):
        if flag is True:
            self.video.startLive()
        else:
            self.video.stopLive()

    """ END Overloading of AbstractVideoDevice methods """


def test_hwo(hwo):
    print(("Image dimensions: %s" % hwo.get_image_dimensions()))
    print(("Live Mode: %s" % hwo.get_video_live()))



import logging
import math

from pydantic import Field

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.ESRF.queue_entry.ssx_base_queue_entry import (
    BaseUserCollectionParameters,
    SsxBaseQueueEntry,
    SsxBaseQueueTaskParameters,
)
from mxcubecore.model.queue_model_objects import DataCollection

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class SsxLineScanCollectionUserParameters(BaseUserCollectionParameters):
    line_range: float = Field(50, gt=0, unit="μm")
    num_images: int = Field(0, gt=0, unit="#")
    spacing: float = Field(10, gt=0, unit="μm")
    # exp_time: float = Field(100e-6, gt=0, lt=1, description='s')

    class Config:
        extra: "ignore"


class SsxLineScanCollectionTaskParameters(SsxBaseQueueTaskParameters):
    user_collection_parameters: SsxLineScanCollectionUserParameters

    @staticmethod
    def update_dependent_fields(field_data):
        # new_data = {"exp_time": field_data["sub_sampling"] * 2}
        return {}


class SsxLineScanCollectionQueueModel(DataCollection):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class SsxLineScanCollectionQueueEntry(SsxBaseQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    QMO = SsxLineScanCollectionQueueModel
    DATA_MODEL = SsxLineScanCollectionTaskParameters
    NAME = "SSX Line Scan"
    REQUIRES = ["point", "line", "no_shape", "chip", "grid"]

    def __init__(self, view, data_model: SsxLineScanCollectionQueueModel):
        super().__init__(view=view, data_model=data_model)

    def execute(self):
        super().execute()

        enforce_centring_phase = False

        motor_x = HWR.beamline.diffractometer.get_object_by_role("ssx_translation")
        motor_y = HWR.beamline.diffractometer.get_object_by_role("phiy")
        motor_z = HWR.beamline.diffractometer.get_object_by_role("focus")

        exp_time = self._data_model._task_data.user_collection_parameters.exp_time
        fname_prefix = self._data_model._task_data.path_parameters.prefix
        num_images = self._data_model._task_data.user_collection_parameters.num_images
        line_range = self._data_model._task_data.user_collection_parameters.line_range
        spacing = self._data_model._task_data.user_collection_parameters.spacing
        sub_sampling = (
            self._data_model._task_data.user_collection_parameters.sub_sampling
        )
        reject_empty_frames = (
            self._data_model._task_data.user_collection_parameters.reject_empty_frames
        )

        self._data_model._task_data.collection_parameters.num_images = num_images
        data_root_path = self.get_data_path()

        num_img_per_rep = line_range // spacing + 1
        num_repetitions = math.ceil(num_images / num_img_per_rep)

        # distance between center and edges (divide by 1000 for μm to mm conversion)
        delta_range = (line_range - (line_range % spacing)) / 2000

        self.take_pedestal()

        HWR.beamline.detector.prepare_acquisition(
            num_images,
            exp_time,
            data_root_path,
            fname_prefix,
            dense_skip_nohits=reject_empty_frames,
        )
        HWR.beamline.detector.wait_ready()

        self.start_processing("LINE-SCAN")

        logging.getLogger("user_level_log").info(f"Preparing scan")

        HWR.beamline.diffractometer.prepare_ssx_line_scan(
            motor_x.get_value() - delta_range,
            motor_y.get_value(),
            motor_z.get_value(),
            motor_x.get_value() + delta_range,
            motor_y.get_value(),
            motor_z.get_value(),
            num_img_per_rep,
            num_repetitions,
        )

        logging.getLogger("user_level_log").info(
            f"Total number of images: {num_images}"
        )
        logging.getLogger("user_level_log").info(f"Images per line: {num_img_per_rep}")
        logging.getLogger("user_level_log").info(
            f"Number of repititions: {num_repetitions}"
        )

        logging.getLogger("user_level_log").info(f"Line range: {delta_range}")
        logging.getLogger("user_level_log").info(
            f"X start: {motor_x.get_value() - delta_range}"
        )
        logging.getLogger("user_level_log").info(
            f"X end: {motor_x.get_value() + delta_range}"
        )

        HWR.beamline.diffractometer.set_phase("DataCollection")
        HWR.beamline.diffractometer.wait_ready()

        if HWR.beamline.control.safshut_oh2.state.name != "OPEN":
            logging.getLogger("user_level_log").info(f"Opening OH2 safety shutter")
            HWR.beamline.control.safshut_oh2.open()

        logging.getLogger("user_level_log").info(f"Acquiring ...")
        HWR.beamline.detector.start_acquisition()
        HWR.beamline.diffractometer.start_ssx_line_scan(enforce_centring_phase)

        logging.getLogger("user_level_log").info(
            f"Waiting for acqusition to finish ..."
        )

        HWR.beamline.diffractometer.wait_ready()
        HWR.beamline.detector.wait_ready()
        logging.getLogger("user_level_log").info(f"Acquired {num_images} images")

    def pre_execute(self):
        super().pre_execute()

    def post_execute(self):
        super().post_execute()

    def stop(self):
        super().stop()



#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging
import time

from mxcubecore import HardwareRepository as HWR
from mxcubecore.queue_entry.base_queue_entry import (
    BaseQueueEntry,
    QueueAbortedException,
)

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class GenericWorkflowQueueEntry(BaseQueueEntry):
    def __init__(self, view=None, data_model=None):
        BaseQueueEntry.__init__(self, view, data_model)
        self.rpc_server_hwobj = None
        self.workflow_running = False
        self.workflow_started = False

    def execute(self):
        BaseQueueEntry.execute(self)

        workflow_hwobj = HWR.beamline.workflow

        # Start execution of a new workflow
        if str(workflow_hwobj.state.value) != "ON":
            # We are trying to start a new workflow and the Tango server is not idle,
            # therefore first abort any running workflow:
            workflow_hwobj.abort()
            if workflow_hwobj.command_failure():
                msg = (
                    "Workflow abort command failed! Please check workflow Tango server."
                )
                logging.getLogger("user_level_log").error(msg)
            else:
                # Then sleep three seconds for allowing the server to abort a running
                # workflow:
                time.sleep(3)
                # If the Tango server has been restarted the state.value is None.
                # If not wait till the state.value is "ON":
                if workflow_hwobj.state.value is not None:
                    while str(workflow_hwobj.state.value) != "ON":
                        time.sleep(0.5)

        msg = "Starting workflow (%s), please wait." % (self.get_data_model()._type)
        logging.getLogger("user_level_log").info(msg)
        workflow_params = self.get_data_model().params_list
        # Add the current node id to workflow parameters
        # group_node_id = self._parent_container._data_model._node_id
        # workflow_params.append("group_node_id")
        # workflow_params.append("%d" % group_node_id)
        workflow_hwobj.start(workflow_params)
        if workflow_hwobj.command_failure():
            msg = "Workflow start command failed! Please check workflow Tango server."
            logging.getLogger("user_level_log").error(msg)
            self.workflow_running = False
        else:
            self.workflow_running = True
            while self.workflow_running:
                time.sleep(1)

    def workflow_state_handler(self, state):
        if isinstance(state, tuple):
            state = str(state[0])
        else:
            state = str(state)

        if state == "ON":
            self.workflow_running = False
        elif state == "RUNNING":
            self.workflow_started = True
        elif state == "OPEN":
            msg = "Workflow waiting for input, verify parameters and press continue."
            logging.getLogger("user_level_log").warning(msg)
            self.get_queue_controller().show_workflow_tab()

    def pre_execute(self):
        BaseQueueEntry.pre_execute(self)
        qc = self.get_queue_controller()
        workflow_hwobj = HWR.beamline.workflow

        qc.connect(workflow_hwobj, "stateChanged", self.workflow_state_handler)

    def post_execute(self):
        BaseQueueEntry.post_execute(self)
        qc = self.get_queue_controller()
        workflow_hwobj = HWR.beamline.workflow
        qc.disconnect(workflow_hwobj, "stateChanged", self.workflow_state_handler)
        # reset state
        self.workflow_started = False
        self.workflow_running = False

        self.get_data_model().set_executed(True)
        self.get_data_model().set_enabled(False)

    def stop(self):
        BaseQueueEntry.stop(self)
        workflow_hwobj = HWR.beamline.workflow
        workflow_hwobj.abort()
        self.get_view().setText(1, "Stopped")
        raise QueueAbortedException("Queue stopped", self)
